# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Types for handling plaintext padding.
"""

__all__ = ["Padding", "PlaintextMeasurements"]


from secrets import token_bytes

from aiootp._typing import Typing as t
from aiootp._constants import DEFAULT_TTL, BIG
from aiootp.asynchs import asleep
from aiootp.commons import FrozenSlots, FrozenInstance


class PlaintextMeasurements(FrozenSlots):
    """
    Efficiently stores plaintext measurements in instance attributes
    which are used to determine the padding that's needed.
    """

    __slots__ = ("padding_size", "pad_sentinel")

    def __init__(self, /, padding_size: int, pad_sentinel: bytes) -> None:
        self.padding_size = padding_size
        self.pad_sentinel = pad_sentinel


class Padding(FrozenInstance):
    """
    Manages the (de-)padding of plaintext with various values which
    improve the salt misuse-reuse resistance, replay attack mitigations,
    & deniability, of the package's ciphers.

     ______________________________________
    |                                      |
    |  Format Diagram:  Plaintext Padding  |
    |______________________________________|
     __________________________________________________________________
    |                      |                      |                    |
    |      Inner-Header    |        Body          |       Footer       |
    |-----------|----------|----------------------|---------|----------|
    | timestamp | SIV-key  |      plaintext       | padding | sentinel |
    |___________|__________|______________________|_________|__________|

    `Inner-Header`: Prepends a timestamp & SIV-key. The timestamp supports
        a time-to-live feature for ciphertexts which can mitigate replay
        attacks. Together with the random SIV-key, the uniqueness of the
        session's initialization is ensured on every tick of the clock,
        therefore extending salt misuse-reuse resistance. The risk
        calculation starts with setting r = len(iv + siv_key) / 3. Then,
        all else staying constant, once 256**r messages are encrypted
        within a second, each additional encrypted message within that
        same second begins to have more than a 256**(-r) chance of
        generating a repeat context.

    `Footer`: The end padding is random bytes & an encoded length up to
        the block size. The randomness of the end padding, its minimal
        corroborability with secret data / session values, & the
        cipher's large & variable effective key-space, aims to aid the
        cipher's deniability.
    """

    __slots__ = ("config",)

    def __init__(self, /, config: t.ConfigType) -> None:
        """
        Populates the instance with the specified configuration.
        """
        self.config = config

    async def _amake_timestamp(self, /) -> bytes:
        """
        Returns a timestamp measured in seconds from the epoch set by
        the package (1672531200: Sun, 01 Jan 2023 00:00:00 UTC).
        """
        c = self.config
        return await c.clock.amake_timestamp(size=c.TIMESTAMP_BYTES)

    def _make_timestamp(self, /) -> bytes:
        """
        Returns a timestamp measured in seconds from the epoch set by
        the package (1672531200: Sun, 01 Jan 2023 00:00:00 UTC).
        """
        c = self.config
        return c.clock.make_timestamp(size=c.TIMESTAMP_BYTES)

    async def _amake_siv_key(self, /) -> bytes:
        """
        Returns a sequence of random bytes. This value is used to ensure
        every encryption is randomized & unique even if the `key`,
        `salt`, `aad`, & `iv` remain static.
        """
        await asleep()
        return token_bytes(self.config.SIV_KEY_BYTES)

    def _make_siv_key(self, /) -> bytes:
        """
        Returns a sequence of random bytes. This value is used to ensure
        every encryption is randomized & unique even if the `key`,
        `salt`, `aad`, & `iv` remain static.
        """
        return token_bytes(self.config.SIV_KEY_BYTES)

    async def astart_padding(self, /) -> bytes:
        """
        Prepends a timestamp & SIV-key. The timestamp supports a
        time-to-live feature for ciphertexts which can mitigate replay
        attacks. Together with the random SIV-key, the uniqueness of the
        session's initialization is ensured on every tick of the clock,
        therefore extending salt misuse-reuse resistance. The risk
        calculation starts with setting r = len(iv + siv_key) / 3. Then,
        all else staying constant, once 256**r messages are encrypted
        within a second, each additional encrypted message within that
        same second begins to have more than a 256**(-r) chance of
        generating a repeat context.
        """
        return await self._amake_timestamp() + await self._amake_siv_key()

    def start_padding(self, /) -> bytes:
        """
        Prepends a timestamp & SIV-key. The timestamp supports a
        time-to-live feature for ciphertexts which can mitigate replay
        attacks. Together with the random SIV-key, the uniqueness of the
        session's initialization is ensured on every tick of the clock,
        therefore extending salt misuse-reuse resistance. The risk
        calculation starts with setting r = len(iv + siv_key) / 3. Then,
        all else staying constant, once 256**r messages are encrypted
        within a second, each additional encrypted message within that
        same second begins to have more than a 256**(-r) chance of
        generating a repeat context.
        """
        return self._make_timestamp() + self._make_siv_key()

    async def _amake_extra_padding(self, /) -> bytes:
        """
        Returns a number of random bytes equal to a positive integer
        multiple of the length of a block. The multiple is `1` by
        default.

        These bytes provide a kind of deniability, where an adversary
        with even unlimited computational capability can't distinguish
        between all possible legitimate plaintexts. This is because
        there are enough random padding degrees of freedom (for large a
        blocksize) for such an adversary to create a super-exponentially
        large number of plaintexts which appear legitimate under any key
        (satisfy verification tags) & appear plausible (they can be made
        to be almost any message + random padding).
        """
        await asleep()
        return token_bytes(self.config.PADDING_FRAME)

    def _make_extra_padding(self, /) -> bytes:
        """
        Returns a number of random bytes equal to a positive integer
        multiple of the length of a block. The multiple is `1` by
        default.

        These bytes provide a kind of deniability, where an adversary
        with even unlimited computational capability can't distinguish
        between all possible legitimate plaintexts. This is because
        there are enough random padding degrees of freedom (for large a
        blocksize) for such an adversary to create a super-exponentially
        large number of plaintexts which appear legitimate under any key
        (satisfy verification tags) & appear plausible (they can be made
        to be almost any message + random padding).
        """
        return token_bytes(self.config.PADDING_FRAME)

    async def _adata_measurements(
        self, /, size: int
    ) -> PlaintextMeasurements:
        """
        Does padding measurements based on the `size` of some unpadded
        data & stores the findings in an object for convenient usage.
        """
        await asleep()
        c = self.config
        remainder = (c.INNER_HEADER_BYTES + size) % c.PADDING_FRAME
        padding_size = c.PADDING_FRAME - remainder
        sentinel = padding_size % c.PADDING_FRAME
        return PlaintextMeasurements(
            padding_size=padding_size,
            pad_sentinel=sentinel.to_bytes(c.SENTINEL_BYTES, BIG),
        )

    def _data_measurements(self, /, size: int) -> PlaintextMeasurements:
        """
        Does padding measurements based on the `size` of some unpadded
        data & stores the findings in an object for convenient usage.
        """
        c = self.config
        remainder = (c.INNER_HEADER_BYTES + size) % c.PADDING_FRAME
        padding_size = c.PADDING_FRAME - remainder
        sentinel = padding_size % c.PADDING_FRAME
        return PlaintextMeasurements(
            padding_size=padding_size,
            pad_sentinel=sentinel.to_bytes(c.SENTINEL_BYTES, BIG),
        )

    async def _amake_end_padding(
        self, /, report: PlaintextMeasurements
    ) -> bytes:
        """
        Returns an excess (frame) of random padding & its encoded length
        to equalize the time it takes to apply the padding even for
        varying plaintext lengths.
        """
        extra_padding = await self._amake_extra_padding()
        return extra_padding + report.pad_sentinel

    def _make_end_padding(self, /, report: PlaintextMeasurements) -> bytes:
        """
        Returns an excess (frame) of random padding & its encoded length
        to equalize the time it takes to apply the padding even for
        varying plaintext lengths.
        """
        extra_padding = self._make_extra_padding()
        return extra_padding + report.pad_sentinel

    async def aend_padding(self, /, size: int) -> bytes:
        """
        The end padding is random bytes & an encoded length up to the
        block size. The randomness of the end padding, its minimal
        corroborability with user secrets / session values, & the
        cipher's large & variable effective key-space, aims to aid the
        cipher's deniability.
        """
        report = await self._adata_measurements(size)
        padding = await self._amake_end_padding(report)
        return padding[-report.padding_size :]

    def end_padding(self, /, size: int) -> bytes:
        """
        The end padding is random bytes & an encoded length up to the
        block size. The randomness of the end padding, its minimal
        corroborability with user secrets / session values, & the
        cipher's large & variable effective key-space, aims to aid the
        cipher's deniability.
        """
        report = self._data_measurements(size)
        padding = self._make_end_padding(report)
        return padding[-report.padding_size :]

    async def apad_plaintext(self, /, data: bytes) -> bytes:
        """
        Pads & returns the plaintext `data` with various values that
        aids a cipher's salt misuse-reuse resistance, replay attack
        mitigations, & deniability.

        Prepends a timestamp & SIV-key. The timestamp supports a
        time-to-live feature for ciphertexts which can mitigate replay
        attacks. Together with the random SIV-key, the uniqueness of the
        session's initialization is ensured on every tick of the clock,
        therefore extending salt misuse-reuse resistance. The risk
        calculation starts with setting r = len(iv + siv_key) / 3. Then,
        all else staying constant, once 256**r messages are encrypted
        within a second, each additional encrypted message within that
        same second begins to have more than a 256**(-r) chance of
        generating a repeat context.

        The end padding is random bytes & an encoded length up to the
        block size. The randomness of the end padding, its minimal
        corroborability with user secrets / session values, & the
        cipher's large & variable effective key-space, aims to aid the
        cipher's deniability.
        """
        start_padding = await self.astart_padding()
        end_padding = await self.aend_padding(len(data))
        return b"".join((start_padding, data, end_padding))

    def pad_plaintext(self, /, data: bytes) -> bytes:
        """
        Pads & returns the plaintext `data` with various values that
        aids a cipher's salt misuse-reuse resistance, replay attack
        mitigations, & deniability.

        Prepends a timestamp & SIV-key. The timestamp supports a
        time-to-live feature for ciphertexts which can mitigate replay
        attacks. Together with the random SIV-key, the uniqueness of the
        session's initialization is ensured on every tick of the clock,
        therefore extending salt misuse-reuse resistance. The risk
        calculation starts with setting r = len(iv + siv_key) / 3. Then,
        all else staying constant, once 256**r messages are encrypted
        within a second, each additional encrypted message within that
        same second begins to have more than a 256**(-r) chance of
        generating a repeat context.

        The end padding is random bytes & an encoded length up to the
        block size. The randomness of the end padding, its minimal
        corroborability with user secrets / session values, & the
        cipher's large & variable effective key-space, aims to aid the
        cipher's deniability.
        """
        start_padding = self.start_padding()
        end_padding = self.end_padding(len(data))
        return b"".join((start_padding, data, end_padding))

    async def adepadding_start_index(self, /) -> int:
        """
        Returns a start index which is used to slice off the prepended
        timestamp & SIV-key from a plaintext.
        """
        return self.config.INNER_HEADER_BYTES

    def depadding_start_index(self, /) -> int:
        """
        Returns a start index which is used to slice off the prepended
        timestamp & SIV-key from a plaintext.
        """
        return self.config.INNER_HEADER_BYTES

    async def adepadding_end_index(self, /, data: bytes) -> int:
        """
        Returns an end index which is used to slice off the appended
        values from some plaintext `data`:
        - The appended random padding.
        - The appended padding sentinel.
        """
        sentinel = int.from_bytes(data[self.config.SENTINEL_SLICE], BIG)
        return -(sentinel if sentinel else self.config.PADDING_FRAME)

    def depadding_end_index(self, /, data: bytes) -> int:
        """
        Returns an end index which is used to slice off the appended
        values from some plaintext `data`:
        - The appended random padding.
        - The appended padding sentinel.
        """
        sentinel = int.from_bytes(data[self.config.SENTINEL_SLICE], BIG)
        return -(sentinel if sentinel else self.config.PADDING_FRAME)

    async def adepad_plaintext(
        self, /, data: bytes, *, ttl: t.Optional[int] = DEFAULT_TTL
    ) -> bytes:
        """
        Returns `data` after these values are removed:
        - The prepended timestamp.
        - The prepended SIV-key.
        - The appended random padding.
        - The appended padding sentinel.
        """
        config = self.config
        config.clock.test_timestamp(data[config.TIMESTAMP_SLICE], ttl=ttl)
        start_index = await self.adepadding_start_index()
        end_index = await self.adepadding_end_index(data)
        return data[start_index:end_index]

    def depad_plaintext(
        self, /, data: bytes, *, ttl: t.Optional[int] = DEFAULT_TTL
    ) -> bytes:
        """
        Returns `data` after these values are removed:
        - The prepended timestamp.
        - The prepended SIV-key.
        - The appended random padding.
        - The appended padding sentinel.
        """
        config = self.config
        config.clock.test_timestamp(data[config.TIMESTAMP_SLICE], ttl=ttl)
        start_index = self.depadding_start_index()
        end_index = self.depadding_end_index(data)
        return data[start_index:end_index]


module_api = dict(
    Padding=t.add_type(Padding),
    PlaintextMeasurements=t.add_type(PlaintextMeasurements),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *


OTHER_TYPE_MAPPING = {Ed25519.__name__: X25519, X25519.__name__: Ed25519}


async def basic_async_tests(tested_class) -> None:
    secret_key_a = await tested_class().agenerate()

    problem = (  # fmt: skip
        "An invalid type was allowed to be imported as a secret key."
    )
    class_not_being_tested = OTHER_TYPE_MAPPING[tested_class.__name__]
    some_secret_key = class_not_being_tested().generate()
    async with Ignore(TypeError, if_else=violation(problem)):
        await tested_class().aimport_secret_key(some_secret_key.secret_key)
    async with Ignore(TypeError, if_else=violation(problem)):
        await tested_class().aimport_secret_key(some_secret_key.public_key)

    problem = (  # fmt: skip
        "An invalid type was allowed to be imported as a public key."
    )
    some_public_key = class_not_being_tested().import_public_key(
        some_secret_key.public_key
    )
    async with Ignore(TypeError, if_else=violation(problem)):
        await tested_class().aimport_secret_key(some_public_key.public_key)

    # Testing equality of async constructors
    key_a_from_secret_object = await tested_class().aimport_secret_key(
        secret_key_a.secret_key
    )
    key_a_from_secret_bytes = await tested_class().aimport_secret_key(
        secret_key_a.secret_bytes
    )

    key_a_from_public_object = await tested_class().aimport_public_key(
        secret_key_a.public_key
    )
    key_a_from_public_bytes = await tested_class().aimport_public_key(
        secret_key_a.public_bytes
    )

    problem = (  # fmt: skip
        "A falsey value secret key import didn't fail."
    )
    async with Ignore(ValueError, if_else=violation(problem)):
        await tested_class().aimport_secret_key(b"")

    problem = (  # fmt: skip
        "A non-key type secret key import didn't fail."
    )
    for non_key_type in (None, "", 0):
        async with Ignore(TypeError, if_else=violation(problem)):
            await tested_class().aimport_secret_key(non_key_type)

    problem = (  # fmt: skip
        "An invalid length secret key import didn't fail."
    )
    for invalid_length in (1, 16, 31, 33, 48, 64):
        async with Ignore(ValueError, if_else=violation(problem)):
            await tested_class().aimport_secret_key(
                token_bytes(invalid_length)
            )

    problem = (  # fmt: skip
        "A falsey value public key import didn't fail."
    )
    async with Ignore(ValueError, if_else=violation(problem)):
        await tested_class().aimport_public_key(b"")

    problem = (  # fmt: skip
        "A non-key type public key import didn't fail."
    )
    for falsey_value in (None, "", 0):
        async with Ignore(TypeError, if_else=violation(problem)):
            await tested_class().aimport_public_key(falsey_value)

    problem = (  # fmt: skip
        "An invalid length public key import didn't fail."
    )
    for invalid_length in (1, 16, 31, 33, 48, 64):
        async with Ignore(ValueError, if_else=violation(problem)):
            await tested_class().aimport_public_key(
                token_bytes(invalid_length)
            )

    assert len(secret_key_a.public_bytes) == 32
    assert len(secret_key_a.secret_bytes) == 32
    assert type(key_a_from_public_object.public_key) is type(
        key_a_from_secret_object.public_key
    )
    assert type(key_a_from_public_object.public_bytes) is bytes
    assert type(key_a_from_public_bytes.public_bytes) is bytes
    assert (
        key_a_from_public_object.public_bytes
        == key_a_from_secret_object.public_bytes
    )
    assert (
        key_a_from_public_bytes.public_bytes
        == key_a_from_secret_bytes.public_bytes
    )
    assert (
        secret_key_a.secret_bytes == key_a_from_secret_object.secret_bytes
    )
    assert secret_key_a.secret_bytes == key_a_from_secret_bytes.secret_bytes
    assert (
        secret_key_a.public_bytes == key_a_from_secret_object.public_bytes
    )
    assert secret_key_a.public_bytes == key_a_from_secret_bytes.public_bytes
    assert (
        secret_key_a.secret_bytes != key_a_from_secret_object.public_bytes
    )
    assert not key_a_from_public_object.has_secret_key()
    assert not key_a_from_public_bytes.has_secret_key()
    assert key_a_from_public_object.has_public_key()
    assert key_a_from_public_bytes.has_public_key()
    assert secret_key_a.has_secret_key()
    assert secret_key_a.has_public_key()

    problem = (  # fmt: skip
        "Async public key import was allowed when an instance was "
        "already initialized with a key."
    )
    async with Ignore(PermissionError, if_else=violation(problem)):
        await secret_key_a.aimport_public_key(secret_key_a.public_bytes)

    problem = (  # fmt: skip
        "Async secret key import was allowed when an instance was "
        "already initialized with a key."
    )
    async with Ignore(PermissionError, if_else=violation(problem)):
        await secret_key_a.aimport_secret_key(secret_key_a.secret_bytes)


def basic_sync_tests(tested_class) -> None:
    secret_key_b = tested_class().generate()

    # Testing equality of sync constructors
    key_b_from_secret_object = tested_class().import_secret_key(
        secret_key_b.secret_key
    )
    key_b_from_secret_bytes = tested_class().import_secret_key(
        secret_key_b.secret_bytes
    )

    key_b_from_public_object = tested_class().import_public_key(
        secret_key_b.public_key
    )
    key_b_from_public_bytes = tested_class().import_public_key(
        secret_key_b.public_bytes
    )

    problem = (  # fmt: skip
        "A falsey value secret key import didn't fail."
    )
    with Ignore(ValueError, if_else=violation(problem)):
        tested_class().import_secret_key(b"")

    problem = (  # fmt: skip
        "A non-key type secret key import didn't fail."
    )
    for non_key_type in (None, "", 0):
        with Ignore(TypeError, if_else=violation(problem)):
            tested_class().import_secret_key(non_key_type)

    problem = (  # fmt: skip
        "An invalid length secret key import didn't fail."
    )
    for invalid_length in (1, 16, 31, 33, 48, 64):
        with Ignore(ValueError, if_else=violation(problem)):
            tested_class().import_secret_key(token_bytes(invalid_length))

    problem = (  # fmt: skip
        "A falsey value public key import didn't fail."
    )
    with Ignore(ValueError, if_else=violation(problem)):
        tested_class().import_public_key(b"")

    problem = (  # fmt: skip
        "A non-key type public key import didn't fail."
    )
    for falsey_value in (None, "", 0):
        with Ignore(TypeError, if_else=violation(problem)):
            tested_class().import_public_key(falsey_value)

    problem = (  # fmt: skip
        "An invalid length public key import didn't fail."
    )
    for invalid_length in (1, 16, 31, 33, 48, 64):
        with Ignore(ValueError, if_else=violation(problem)):
            tested_class().import_public_key(token_bytes(invalid_length))

    assert len(secret_key_b.public_bytes) == 32
    assert len(secret_key_b.secret_bytes) == 32
    assert type(key_b_from_public_object.public_key) is type(
        key_b_from_secret_object.public_key
    )
    assert type(key_b_from_public_object.public_bytes) is bytes
    assert type(key_b_from_public_bytes.public_bytes) is bytes
    assert (
        key_b_from_public_object.public_bytes
        == key_b_from_secret_object.public_bytes
    )
    assert (
        key_b_from_public_bytes.public_bytes
        == key_b_from_secret_bytes.public_bytes
    )
    assert (
        secret_key_b.secret_bytes == key_b_from_secret_object.secret_bytes
    )
    assert secret_key_b.secret_bytes == key_b_from_secret_bytes.secret_bytes
    assert (
        secret_key_b.public_bytes == key_b_from_secret_object.public_bytes
    )
    assert secret_key_b.public_bytes == key_b_from_secret_bytes.public_bytes
    assert (
        secret_key_b.secret_bytes != key_b_from_secret_object.public_bytes
    )
    assert not key_b_from_public_object.has_secret_key()
    assert not key_b_from_public_bytes.has_secret_key()
    assert key_b_from_public_object.has_public_key()
    assert key_b_from_public_bytes.has_public_key()
    assert secret_key_b.has_secret_key()
    assert secret_key_b.has_public_key()

    problem = (  # fmt: skip
        "Sync public key import was allowed when an instance was "
        "already initialized with a key."
    )
    with Ignore(PermissionError, if_else=violation(problem)):
        secret_key_b.import_public_key(secret_key_b.public_bytes)

    problem = (  # fmt: skip
        "Sync secret key import was allowed when an instance was "
        "already initialized with a key."
    )
    with Ignore(PermissionError, if_else=violation(problem)):
        secret_key_b.import_secret_key(secret_key_b.secret_bytes)


async def test_X25519() -> None:
    # Test class constructors
    secret_key_a = await X25519().agenerate()
    secret_key_b = X25519().generate()

    await basic_async_tests(X25519)
    basic_sync_tests(X25519)

    # exchange methods create shared keys from different instances'
    # public bytes
    shared_key_a = secret_key_a.exchange(secret_key_b.public_bytes)
    shared_key_b = secret_key_b.exchange(secret_key_a.public_bytes)
    assert bytes_are_equal(shared_key_a, shared_key_b)

    shared_key_a = await secret_key_a.aexchange(secret_key_b.public_bytes)
    shared_key_b = await secret_key_b.aexchange(secret_key_a.public_bytes)
    assert await abytes_are_equal(shared_key_a, shared_key_b)

    # exchange methods create shared keys from different instances'
    # public key object
    shared_key_a = secret_key_a.exchange(secret_key_b.public_key)
    shared_key_b = secret_key_b.exchange(secret_key_a.public_key)
    assert bytes_are_equal(shared_key_a, shared_key_b)

    shared_key_a = await secret_key_a.aexchange(secret_key_b.public_key)
    shared_key_b = await secret_key_b.aexchange(secret_key_a.public_key)
    assert await abytes_are_equal(shared_key_a, shared_key_b)

    # exchange methods create shared keys from different instances'
    # secret key object
    problem = (  # fmt: skip
        "A secret key was used in an interface expecting a public key."
    )
    with Ignore(TypeError, if_else=violation(problem)):
        secret_key_a.exchange(secret_key_b.secret_key)
    with Ignore(TypeError, if_else=violation(problem)):
        secret_key_b.exchange(secret_key_a.secret_key)

    with Ignore(TypeError, if_else=violation(problem)):
        await secret_key_a.aexchange(secret_key_b.secret_key)
    with Ignore(TypeError, if_else=violation(problem)):
        await secret_key_b.aexchange(secret_key_a.secret_key)


class TestDiffieHellmanProtocols:
    async def test_non_kex_types_throw_error_in_kex_protocol_inits(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "A non-key exchange type was supplied during init & didn't fail."
        )
        for non_kex_type in (dict, str, int, bytes, t.Namespace, Ed25519):
            async with Ignore(TypeError, if_else=violation(problem)):
                t.DoubleDiffieHellmanClient(
                    non_kex_type, kdf_type=DomainKDF
                )
            async with Ignore(TypeError, if_else=violation(problem)):
                t.DoubleDiffieHellmanServer(
                    non_kex_type(), kdf_type=DomainKDF
                )
            async with Ignore(TypeError, if_else=violation(problem)):
                t.TripleDiffieHellmanClient(
                    non_kex_type(), kdf_type=DomainKDF
                )
            async with Ignore(TypeError, if_else=violation(problem)):
                t.TripleDiffieHellmanServer(
                    non_kex_type(), kdf_type=DomainKDF
                )

    async def test_non_kdf_types_throw_error_in_kex_protocol_inits(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "A non-KDF type was supplied during init & didn't fail."
        )
        my_identity_key = X25519().generate()
        for non_kdf_type in (dict, str, int, bytes, t.Namespace, Ed25519):
            async with Ignore(TypeError, if_else=violation(problem)):
                t.DoubleDiffieHellmanClient(X25519, kdf_type=non_kdf_type)
            async with Ignore(TypeError, if_else=violation(problem)):
                t.DoubleDiffieHellmanServer(
                    my_identity_key, kdf_type=non_kdf_type
                )
            async with Ignore(TypeError, if_else=violation(problem)):
                t.TripleDiffieHellmanClient(
                    my_identity_key, kdf_type=non_kdf_type
                )
            async with Ignore(TypeError, if_else=violation(problem)):
                t.TripleDiffieHellmanServer(
                    my_identity_key, kdf_type=non_kdf_type
                )

    async def test_async_double_diffie_hellman(self) -> None:
        server_key = await X25519().agenerate()
        client = X25519.dh2_client()
        server = server_key.dh2_server()

        client_ephemeral_key = await client.asend(server_key.public_bytes)
        server_kdf = await server.areceive(client_ephemeral_key)
        server_ephemeral_key = await server.asend()
        client_kdf = await client.areceive(server_ephemeral_key)
        assert client_kdf.sha3_512() == server_kdf.sha3_512()

    async def test_sync_double_diffie_hellman(self) -> None:
        server_key = X25519().generate()
        client = X25519.dh2_client()
        server = server_key.dh2_server()

        client_ephemeral_key = client.send(server_key.public_bytes)
        server_kdf = server.receive(client_ephemeral_key)
        server_ephemeral_key = server.send()
        client_kdf = client.receive(server_ephemeral_key)
        assert client_kdf.sha3_512() == server_kdf.sha3_512()

    async def test_async_triple_diffie_hellman(self) -> None:
        client_key = await X25519().agenerate()
        server_key = await X25519().agenerate()
        client = client_key.dh3_client()
        server = server_key.dh3_server()

        client_identity_key, client_ephemeral_key = await client.asend(
            server_key.public_bytes
        )
        server_kdf = await server.areceive(
            client_identity_key, client_ephemeral_key
        )
        server_ephemeral_key = await server.asend()
        client_kdf = await client.areceive(server_ephemeral_key)
        assert client_kdf.sha3_512() == server_kdf.sha3_512()

    async def test_sync_triple_diffie_hellman(self) -> None:
        client_key = X25519().generate()
        server_key = X25519().generate()
        client = client_key.dh3_client()
        server = server_key.dh3_server()

        client_identity_key, client_ephemeral_key = client.send(
            server_key.public_bytes
        )
        server_kdf = server.receive(
            client_identity_key, client_ephemeral_key
        )
        server_ephemeral_key = server.send()
        client_kdf = client.receive(server_ephemeral_key)
        assert client_kdf.sha3_512() == server_kdf.sha3_512()

    async def test_double_diffie_hellman_interop(self) -> None:
        server_key = await X25519().agenerate()
        client = X25519.dh2_client()
        server = server_key.dh2_server()

        client_ephemeral_key = client.send(server_key.public_bytes)
        server_kdf = await server.areceive(client_ephemeral_key)
        server_ephemeral_key = await server.asend()
        client_kdf = client.receive(server_ephemeral_key)
        assert client_kdf.sha3_512() == server_kdf.sha3_512()

        client = X25519.dh2_client()
        server = server_key.dh2_server()

        client_ephemeral_key = await client.asend(server_key.public_bytes)
        server_kdf = server.receive(client_ephemeral_key)
        server_ephemeral_key = server.send()
        client_kdf = await client.areceive(server_ephemeral_key)
        assert client_kdf.sha3_512() == server_kdf.sha3_512()

    async def test_triple_diffie_hellman_interop(self) -> None:
        client_key = await X25519().agenerate()
        server_key = await X25519().agenerate()
        client = client_key.dh3_client()
        server = server_key.dh3_server()

        client_identity_key, client_ephemeral_key = client.send(
            server_key.public_bytes
        )
        server_kdf = await server.areceive(
            client_identity_key, client_ephemeral_key
        )
        server_ephemeral_key = await server.asend()
        client_kdf = client.receive(server_ephemeral_key)
        assert client_kdf.sha3_512() == server_kdf.sha3_512()

        client = client_key.dh3_client()
        server = server_key.dh3_server()

        client_identity_key, client_ephemeral_key = await client.asend(
            server_key.public_bytes
        )
        server_kdf = server.receive(
            client_identity_key, client_ephemeral_key
        )
        server_ephemeral_key = server.send()
        client_kdf = await client.areceive(server_ephemeral_key)
        assert client_kdf.sha3_512() == server_kdf.sha3_512()


async def test_Ed25519() -> None:
    # Test class constructors
    secret_key_a = await Ed25519().agenerate()
    secret_key_b = Ed25519().generate()

    await basic_async_tests(Ed25519)
    basic_sync_tests(Ed25519)

    async_signature = await secret_key_a.asign(plaintext_bytes)
    signature = secret_key_b.sign(plaintext_bytes)

    arbitrary_verifier = Ed25519().generate()
    key_a_verifier = await Ed25519().aimport_public_key(
        secret_key_a.public_bytes
    )
    key_b_verifier = Ed25519().import_public_key(secret_key_b.public_bytes)

    # async verification succeeds when supplied a correct signature & data
    await key_a_verifier.averify(async_signature, plaintext_bytes)
    await arbitrary_verifier.averify(
        async_signature,
        plaintext_bytes,
        public_key=secret_key_a.public_bytes,
    )

    # async verification succeeds when supplied an incorrect signature & data
    problem = (  # fmt: skip
        "Async verification succeeded for an invalid signature."
    )
    async with Ignore(Ed25519.InvalidSignature, if_else=violation(problem)):
        await key_a_verifier.averify(
            token_bytes(len(async_signature)), plaintext_bytes
        )
    problem = (  # fmt: skip
        "Async verification succeeded for an invalid signature."
    )
    async with Ignore(Ed25519.InvalidSignature, if_else=violation(problem)):
        await arbitrary_verifier.averify(
            token_bytes(len(async_signature)),
            plaintext_bytes,
            public_key=secret_key_a.public_bytes,
        )

    # sync verification succeeds when supplied a correct signature & data
    key_b_verifier.verify(signature, plaintext_bytes)
    arbitrary_verifier.verify(
        signature, plaintext_bytes, public_key=secret_key_b.public_bytes
    )

    # sync verification succeeds when supplied an incorrect signature & data
    problem = (  # fmt: skip
        "Verification succeeded for an invalid signature."
    )
    with Ignore(Ed25519.InvalidSignature, if_else=violation(problem)):
        key_a_verifier.verify(
            token_bytes(len(async_signature)), plaintext_bytes
        )
    problem = (  # fmt: skip
        "Verification succeeded for an invalid signature."
    )
    with Ignore(Ed25519.InvalidSignature, if_else=violation(problem)):
        arbitrary_verifier.verify(
            token_bytes(len(async_signature)),
            plaintext_bytes,
            public_key=secret_key_a.public_bytes,
        )


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Wraps file operations from the `os` module in a decorator that runs
those methods in an async executor. This was adapted from the
`aiofiles` package:

https://github.com/Tinche/aiofiles/blob/master/aiofiles/os.py

Whose license is Apache License 2.0, available here:

http://www.apache.org/licenses/LICENSE-2.0
"""

__all__ = [
    "chmod",
    "chown",
    "makedirs",
    "mkdir",
    "rename",
    "remove",
    "rmdir",
    "sendfile",
    "stat",
]


import os
import warnings

from aiootp._typing import Typing as t

from .loops import asleep, wrap_in_executor


async def not_implemented_placeholder(*_: t.Any, **__: t.Any) -> None:
    # fmt: off
    await asleep()                                  # pragma: no cover
    warnings.warn("Function not supported by OS.")  # pragma: no cover
    # fmt: on


defs = {}


for name in __all__:
    if hasattr(os, name):
        defs[name] = wrap_in_executor(getattr(os, name))
    else:
        defs[name] = not_implemented_placeholder  # pragma: no cover


globals().update(defs)


module_api = dict(
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    **defs,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *


class TestConcurrentExecution:
    async def test_tasks_run_in_background(self) -> None:
        async def increment_over_time(track: int) -> None:
            nonlocal value_0, value_1

            if track == 0:
                while value_0 < 16:
                    value_0 += 1
                    await asleep()
            elif track == 1:
                while value_1 < 16:
                    value_1 += 1
                    await asleep()

        value_0 = 0
        value_1 = 0
        asynchs.new_task(increment_over_time(track=0))
        await asleep(0.0001)
        assert 0 < value_0
        assert 0 == value_1
        asynchs.new_task(increment_over_time(track=1))
        await asleep(0.0001)
        assert 1 < value_0
        assert 0 < value_1

        await asleep(0.0001)
        assert 2 < value_0
        assert 1 < value_1

    async def test_futures_run_in_background(self) -> None:
        async def increment_over_time(track: int) -> None:
            nonlocal value_0, value_1

            if track == 0:
                while value_0 < 16:
                    value_0 += 1
                    await asleep()
            elif track == 1:
                while value_1 < 16:
                    value_1 += 1
                    await asleep()

        value_0 = 0
        value_1 = 0
        asynchs.new_future(increment_over_time(track=0))
        await asleep(0.0001)
        assert 0 < value_0
        assert 0 == value_1
        asynchs.new_future(increment_over_time(track=1))
        await asleep(0.0001)
        assert 1 < value_0
        assert 0 < value_1

        await asleep(0.0001)
        assert 2 < value_0
        assert 1 < value_1


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Implements synchronous & asynchronous transparently encrypted data
persistance classes.
"""

__all__ = ["AsyncDatabase", "Database"]


from .dbdomains import *
from .dbkdf import *
from .profile_tokens import *
from .database_properties import *
from .async_database import *
from .sync_database import *


modules = dict(
    database_properties=database_properties,
    dbdomains=dbdomains,
    dbkdf=dbkdf,
    profile_tokens=profile_tokens,
    async_database=async_database,
    sync_database=sync_database,
)


module_api = dict(
    AsyncDatabase=AsyncDatabase,
    Database=Database,
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Asynchrony event loop tools.
"""

__all__ = [
    "AsyncInit",
    "asleep",
    "asyncio",
    "gather",
    "get_event_loop",
    "get_event_loop_id",
    "new_event_loop",
    "new_future",
    "new_task",
    "run",
    "set_event_loop",
    "sleep",
    "wrap_in_executor",
]


import asyncio
from time import sleep
from functools import wraps, partial
from asyncio import sleep as _asleep
from asyncio import run, set_event_loop

from aiootp._typing import Typing as t
from aiootp._debug_control import DebugControl


gather = asyncio.gather
get_event_loop = asyncio.get_event_loop
get_event_loop_id = lambda: id(get_event_loop())
new_event_loop = asyncio.new_event_loop
new_future = asyncio.ensure_future


# Can toggle asyncio's debug mode using DebugControl.enable_debugging()
# WARNING: This will also reveal potentially sensitive values in object
# repr's that are omitted by default.
DebugControl._switches.append(
    lambda: get_event_loop().set_debug(DebugControl.is_debugging())
)


async def asleep(seconds: t.PositiveRealNumber = 0) -> None:
    """
    Async sleep from asyncio.
    """
    await _asleep(seconds)


def new_task(coro: t.Awaitable) -> asyncio.Task:
    """
    Proxy's access to `asyncio.get_event_loop().create_task()`.
    """
    return get_event_loop().create_task(coro)


def wrap_in_executor(
    function: t.Callable[..., t.Any],
) -> t.Coroutine[t.Any, t.Any, t.Any]:
    """
    A decorator that wraps synchronous blocking IO functions so they
    will run in an async executor.
    """

    @wraps(function)
    async def runner(*args: t.Any, **kwargs: t.Any) -> t.Any:
        partial_function = partial(function, *args, **kwargs)
        return await get_event_loop().run_in_executor(
            executor=None, func=partial_function
        )

    return runner


class AsyncInit(type):
    """
    A metaclass which allows classes to use asynchronous `__init__`
    methods. Inspired by David Beazley.
    """

    async def __call__(cls, *a: t.Any, **kw: t.Any) -> t.Self:
        self = cls.__new__(cls, *a, **kw)
        await self.__init__(*a, **kw)
        return self


module_api = dict(
    AsyncInit=t.add_type(AsyncInit),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    asleep=asleep,
    asyncio=asyncio,
    gather=gather,
    get_event_loop=get_event_loop,
    get_event_loop_id=get_event_loop_id,
    new_event_loop=new_event_loop,
    new_future=new_future,
    new_task=new_task,
    run=run,
    set_event_loop=set_event_loop,
    sleep=sleep,
    wrap_in_executor=wrap_in_executor,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
An interface for verfying the authenticity of software package
hashes & signatures.
"""

__all__ = ["PackageVerifier"]


import json
from pathlib import Path
from hashlib import sha384

from aiootp._typing import Typing as t
from aiootp._constants import CHECKSUM, CHECKSUMS, SIGNATURE, SIGNING_KEY
from aiootp._exceptions import Issue, PackageSignerIssue
from aiootp.generics import bytes_are_equal

from .curve_25519 import Ed25519


class PackageVerifier:
    """
    Provides an intuitive API for verifying package summaries produced
    by `PackageSigner` objects.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    from aiootp import PackageVerifier

    verifier = PackageVerifier(public_signing_key, path="")
    verifier.verify_summary(package_signature_summary)
    """

    __slots__ = (
        "_path",
        "_checksum",
        "_signature",
        "_signing_key",
        "_summary_dictionary",
    )

    _Hasher: t.Callable[[bytes], t.HasherType] = sha384
    _Signer: type = Ed25519

    _CHECKSUM: str = CHECKSUM
    _CHECKSUMS: str = CHECKSUMS
    _SIGNATURE: str = SIGNATURE
    _SIGNING_KEY: str = SIGNING_KEY

    InvalidSignature: type = PackageSignerIssue.InvalidSignature

    def __init__(
        self,
        public_signing_key: t.Union[bytes, t.PublicKeyType, t.SignerType],
        *,
        path: t.OptionalPathStr = None,
        verify_files: bool = True,
    ) -> None:
        """
        Receives the bytes type public signing key a user expects a
        package to be signed by, & stores it within the instance. The
        `path` keyword argument is the root location where all of the
        source files can be reached via the relative paths of the files
        declared in the summary. If instead the source files will not be
        checked for validity & only the validity of the signature will
        be assertained, the `verify_files` keyword can be set falsey.
        """
        if public_signing_key.__class__ is not self._Signer:
            key = public_signing_key
            self._signing_key = self._Signer().import_public_key(key)
        else:
            self._signing_key = public_signing_key
        if verify_files:
            self._path = Path(path).absolute()
        elif path is not None:
            raise Issue.unused_parameters("path", "not verifying files")

    @property
    def _summary_bytes(self) -> bytes:
        """
        Returns the UTF-8 encoded JSON package signature summary sans
        the package checksum & signature for hashing.
        """
        return json.dumps(self._summary_dictionary).encode()

    def _import_summary(
        self, summary: t.Dict[str, t.JSONSerializable]
    ) -> None:
        """
        Verifies the package summary checksum & stores its values within
        the instance.
        """
        summary = self._summary_dictionary = {**summary}
        self._checksum = bytes.fromhex(summary.pop(self._CHECKSUM))
        self._signature = bytes.fromhex(summary.pop(self._SIGNATURE))
        if self._Hasher(self._summary_bytes).digest() != self._checksum:
            raise Issue.invalid_value("package summary checksum")

    def _verify_file_checksums(self, summary: dict) -> None:
        """
        Verifies the files declared in the summary by loading them from
        the filesystem & calculating their digests for a match.
        """
        path = self._path
        files = summary[self._CHECKSUMS]
        for file_path, purported_hexdigest in files.items():
            purported_digest = bytes.fromhex(purported_hexdigest)
            with (path / file_path).open("rb") as source_file:
                digest = self._Hasher(source_file.read()).digest()
                if not bytes_are_equal(purported_digest, digest):
                    raise PackageSignerIssue.invalid_file_digest(file_path)

    def verify_summary(
        self,
        summary: t.Union[
            t.Dict[str, t.JSONSerializable], t.JSONDeserializable
        ],
    ) -> None:
        """
        Verifies the purported checksum of a package summary & the
        signature of the checksum.
        """
        if summary.__class__ is not dict:
            summary = json.loads(summary)
        purported_signing_key = bytes.fromhex(summary[self._SIGNING_KEY])
        if purported_signing_key != self._signing_key.public_bytes:
            raise Issue.invalid_value("summary's public signing key")
        self._import_summary(summary)
        if hasattr(self, "_path"):
            self._verify_file_checksums(summary)
        self._signing_key.verify(self._signature, self._checksum)


module_api = dict(
    PackageVerifier=t.add_type(PackageVerifier),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Canonicalized keyed hashing functions for SHA3 objects.
"""

__all__ = ["ahash_bytes", "hash_bytes"]


from hashlib import sha3_512

from aiootp._typing import Typing as t
from aiootp._constants import INT_BYTES, BIG

from .canon import aencode_key, encode_key
from .canon import acanonical_pack, canonical_pack


async def ahash_bytes(
    *collection: bytes,
    hasher: t.Callable[[bytes], t.HasherType] = sha3_512,
    pad: bytes = b"\x00",
    size: t.Optional[int] = None,
    key: bytes = b"",
) -> bytes:
    """
    Joins the `collection` of `bytes`-type objects with a canonical
    encoding & returns the `hasher` object's digest of the encoded
    result.

    `size` may be specified if the `hasher` object's `digest`
    method so requires.

    Returns a keyed-hash if `key` is specified.
    """
    obj = hasher()
    obj.update(
        (await aencode_key(key, obj.block_size, pad=pad) if key else b"")
        + await acanonical_pack(
            (size if size else obj.digest_size).to_bytes(INT_BYTES, BIG),
            *collection,
            blocksize=obj.block_size,
            pad=pad,
        )
    )
    if size:
        return obj.digest(size)
    return obj.digest()


def hash_bytes(
    *collection: bytes,
    hasher: t.Callable[[bytes], t.HasherType] = sha3_512,
    pad: bytes = b"\x00",
    size: t.Optional[int] = None,
    key: bytes = b"",
) -> bytes:
    """
    Joins the `collection` of `bytes`-type objects with a canonical
    encoding & returns the `hasher` object's digest of the encoded
    result.

    `size` may be specified if the `hasher` object's `digest`
    method so requires.

    Returns a keyed-hash if `key` is specified.
    """
    obj = hasher()
    obj.update(
        (encode_key(key, obj.block_size, pad=pad) if key else b"")
        + canonical_pack(
            (size if size else obj.digest_size).to_bytes(INT_BYTES, BIG),
            *collection,
            blocksize=obj.block_size,
            pad=pad,
        )
    )
    if size:
        return obj.digest(size)
    return obj.digest()


module_api = dict(
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    ahash_bytes=ahash_bytes,
    hash_bytes=hash_bytes,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
The server side of a key exchange protocol which can provide
client-side & server-side identity authentication with ephemeral
random secrets.
"""

__all__ = ["TripleDiffieHellmanServer"]


from aiootp._typing import Typing as t
from aiootp._exceptions import Issue
from aiootp.commons import FrozenInstance
from aiootp.generics import Domains


class TripleDiffieHellmanServer(FrozenInstance):
    """
    The server side of a key exchange protocol which can provide
    client-side & server-side identity authentication with ephemeral
    random secrets.

     _____________________________________
    |                                     |
    |          Protocol Diagram:          |
    |_____________________________________|

            -----------------          |         -----------------
            |  Client-side  |          |         |  Server-side  |
            -----------------          |         -----------------
                                       |
    key = X25519().generate()          |         X25519().generate() = key
                                       |
    client = key.dh3_client()          |           key.public_bytes = id_s
                                       |
    id_c, eph_c = client.send(id_s) ------>
                                       |
                                       |         key.dh3_server() = server
                                       |
                                       | server.receive(id_c, eph_c) = kdf
                                       |
                                    <------          server.send() = eph_s
                                       |
    kdf = client.receive(eph_s)        |
                                       |
    ------------------------------------------------------------------------
    """

    __slots__ = (
        "_kdf_type",
        "_key_exchange_type",
        "_my_identity_key",
        "_my_ephemeral_key",
        "_peer_identity_key",
        "_peer_ephemeral_key",
        "_sanitize",
    )

    def __init__(
        self, /, my_identity_key: t.KeyExchangeType, *, kdf_type: type
    ) -> None:
        if not issubclass(kdf_type, t.DomainKDFType):
            raise Issue.value_must_be_subtype("KDF type", t.DomainKDFType)
        elif not isinstance(my_identity_key, t.KeyExchangeType):
            raise Issue.value_must_be_type("KEX ID key", t.KeyExchangeType)
        self._kdf_type = kdf_type
        self._key_exchange_type = my_identity_key.__class__
        self._my_identity_key = my_identity_key
        self._sanitize = self._key_exchange_type._process_public_key

    async def areceive(
        self,
        /,
        peer_identity_key: t.Union[bytes, t.PublicKeyType],
        peer_ephemeral_key: t.Union[bytes, t.PublicKeyType],
    ) -> t.DomainKDFType:
        """
        Receives the identity & ephemeral public keys from an intended
        client, & returns the KDF object which has been primed with the
        public & secret values involed in the exchange.
        """
        my_identity_key = self._my_identity_key
        my_ephemeral_key = self._my_ephemeral_key = (
            await self._key_exchange_type().agenerate()  # fmt: skip
        )
        peer_identity_key = self._peer_identity_key = self._sanitize(
            peer_identity_key
        )
        peer_ephemeral_key = self._peer_ephemeral_key = self._sanitize(
            peer_ephemeral_key
        )
        shared_key_ad = await my_identity_key.aexchange(peer_ephemeral_key)
        shared_key_bc = await my_ephemeral_key.aexchange(peer_identity_key)
        shared_key_cd = await my_ephemeral_key.aexchange(peer_ephemeral_key)
        return self._kdf_type(
            Domains.DH3,
            peer_identity_key.public_bytes_raw(),
            peer_ephemeral_key.public_bytes_raw(),
            my_identity_key.public_bytes,
            my_ephemeral_key.public_bytes,
            key=shared_key_ad + shared_key_bc + shared_key_cd,
        )

    def receive(
        self,
        /,
        peer_identity_key: t.Union[bytes, t.PublicKeyType],
        peer_ephemeral_key: t.Union[bytes, t.PublicKeyType],
    ) -> t.DomainKDFType:
        """
        Receives the identity & ephemeral public keys from an intended
        client, & returns the KDF object which has been primed with the
        public & secret values involed in the exchange.
        """
        my_identity_key = self._my_identity_key
        my_ephemeral_key = self._my_ephemeral_key = (
            self._key_exchange_type().generate()
        )
        peer_identity_key = self._peer_identity_key = self._sanitize(
            peer_identity_key
        )
        peer_ephemeral_key = self._peer_ephemeral_key = self._sanitize(
            peer_ephemeral_key
        )
        shared_key_ad = my_identity_key.exchange(peer_ephemeral_key)
        shared_key_bc = my_ephemeral_key.exchange(peer_identity_key)
        shared_key_cd = my_ephemeral_key.exchange(peer_ephemeral_key)
        return self._kdf_type(
            Domains.DH3,
            peer_identity_key.public_bytes_raw(),
            peer_ephemeral_key.public_bytes_raw(),
            my_identity_key.public_bytes,
            my_ephemeral_key.public_bytes,
            key=shared_key_ad + shared_key_bc + shared_key_cd,
        )

    async def asend(self, /) -> bytes:
        """
        Returns the instance's ephemeral public key involved in the
        exchange to be sent to the client.
        """
        return self._my_ephemeral_key.public_bytes

    def send(self, /) -> bytes:
        """
        Returns the instance's ephemeral public key involved in the
        exchange to be sent to the client.
        """
        return self._my_ephemeral_key.public_bytes


module_api = dict(
    TripleDiffieHellmanServer=t.add_type(TripleDiffieHellmanServer),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A general interface for multi-threading & multi-processing.
"""

__all__ = ["ConcurrencyGuard", "ConcurrencyInterface"]


from hmac import compare_digest
from secrets import token_bytes

from aiootp._typing import Typing as t
from aiootp._exceptions import Issue, IncoherentConcurrencyState
from aiootp.commons import FrozenTypedSlots

from . import is_async_function
from .loops import asleep, sleep, new_event_loop


def process_probe_delay(cls: t.Any, value: t.Optional[float], /) -> float:
    """
    Ensures the probe frequency is positive & returns it, if it's
    specified. Otherwise, returns `cls`'s default value.
    """
    if value is None:
        return cls._default_probe_delay
    elif value > 0:
        return value
    else:
        raise Issue.value_must("probe_delay", "be > 0")


class ConcurrencyGuard(FrozenTypedSlots):
    """
    An interface for queueing execution contexts given only a shared
    `deque` or `deque`-like double-ended queue. Prevents simultaneous /
    out of order runs of blocks of code. A `deque` is recommended since
    it supports atomic operations. Any atomic, shared datastructure with
    `append`, `popleft`, & `queue[0]` methods would fit the API.

     _____________________________________
    |                                     |
    |         Example As Diagram:         |
    |_____________________________________|

                           ----------------------
                           |   Shared Context   |
                           ----------------------

                               queue = deque()
                           -----------------------
            -----------------         |        -----------------
            |   Context A   |         |        |   Context B   |
            -----------------         |        -----------------
                                      |
    with ConcurrencyGuard(queue):     |  with ConcurrencyGuard(queue):
        mutable_thing[0] = 0          |      assert mutable_thing[0] == "done."
        ...                           |      mutable_thing[0] = 1
        ...                           |
        assert mutable_thing[0] == 0  |
        mutable_thing[0] = "done."    |
                                      |
    ----------------------------------------------------------------------
    Context A is called first, but Context B is called before it finishes.
    ----------------------------------------------------------------------
    """

    __slots__ = ("probe_delay", "queue", "token")

    slots_types: t.Mapping[str, type] = dict(
        probe_delay=float,
        queue=t.SupportsAppendPopleft,
        token=bytes,
    )

    _default_probe_delay: float = 0.00001

    IncoherentConcurrencyState: type = IncoherentConcurrencyState

    def __init__(
        self,
        /,
        queue: t.SupportsAppendPopleft,
        *,
        probe_delay: t.Optional[float] = None,
        token: t.Optional[bytes] = None,
    ) -> None:
        """
        `queue`: Atomic, `deque`-like datastructure that supports `append`,
                `popleft`, & `queue[0]` methods.

        `probe_delay`: The float/fractional number of seconds to wait
                before each attempt to detect if the instance's token
                has been authorized to run.

        `token`: The unique authorization token held by this context.
        """
        self.probe_delay = process_probe_delay(self, probe_delay)
        self.queue = queue
        self.token = token_bytes(32) if token is None else token

    async def __aenter__(self, /) -> t.Self:
        """
        Prevents entering the context by asynchronously sleeping until
        the instance's unique authorization token is the current token
        in the 0th position of the queue.
        """
        self.queue.append(self.token)
        while not compare_digest(self.token, self.queue[0]):
            await asleep(self.probe_delay)
        return self

    def __enter__(self, /) -> t.Self:
        """
        Prevents entering the context by synchronously sleeping until
        the instance's unique authorization token is the current token
        in the 0th position of the queue.
        """
        self.queue.append(self.token)
        while not compare_digest(self.token, self.queue[0]):
            sleep(self.probe_delay)
        return self

    async def __aexit__(
        self,
        /,
        exc_type: t.Optional[type] = None,
        exc_value: t.Optional[Exception] = None,
        traceback: t.Optional[t.TracebackType] = None,
    ) -> bool:
        """
        Raises `self.IncoherentConcurrencyState` if another instance's
        authorization token has taken this instance's place in the token
        queue.

        Otherwise, raises any exception raised in the context's code
        block.

        Otherwise, closes the context silently.
        """
        await asleep()
        if not compare_digest(self.token, self.queue.popleft()):
            raise self.IncoherentConcurrencyState from exc_value
        return exc_type is None

    def __exit__(
        self,
        /,
        exc_type: t.Optional[type] = None,
        exc_value: t.Optional[Exception] = None,
        traceback: t.Optional[t.TracebackType] = None,
    ) -> bool:
        """
        Raises `self.IncoherentConcurrencyState` if another instance's
        authorization token has taken this instance's place in the token
        queue.

        Otherwise, raises any exception raised in the context's code
        block.

        Otherwise, closes the context silently.
        """
        if not compare_digest(self.token, self.queue.popleft()):
            raise self.IncoherentConcurrencyState from exc_value
        return exc_type is None


class ConcurrencyInterface:
    """
    Defines an interface for managing tasks & pools of threads & processes.
    """

    __slots__ = ()

    _Manager: type

    _default_probe_delay: t.PositiveRealNumber
    _pool: t.PoolExecutorType
    _type: type

    BrokenPool: type

    @staticmethod
    def _arun_func(
        func: t.Callable[..., t.Any],
        state: t.Sequence[t.Any],
        /,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> None:
        """
        Used by the class to retrieve return values from an async or
        sync `func` run in a new process / thread by storing the result
        in a shared `state` container.
        """
        if is_async_function(func):
            run = new_event_loop().run_until_complete
            state.append(run(func(*args, **kwargs)))
        else:
            state.append(func(*args, **kwargs))

    @staticmethod
    def _run_func(
        func: t.Callable[..., t.Any],
        state: t.Sequence[t.Any],
        /,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> None:
        """
        Used by the class to retrieve return values from a sync `func`
        run in a new process / thread by storing the result in a shared
        `state` container.
        """
        state.append(func(*args, **kwargs))

    @classmethod
    async def anew(
        cls,
        func: t.Callable[..., t.Any],
        /,
        *args: t.Any,
        probe_delay: t.Optional[t.PositiveRealNumber] = None,
        **kwargs: t.Any,
    ) -> t.Any:
        """
        Runs an async or sync function in another process or thread
        depending on the class calling this method, so that heavy cpu-
        bound computations, or blocking IO operations, can better
        coexist with asynchronous code.
        """
        delay = process_probe_delay(cls, probe_delay)
        state = cls._Manager().list()
        task = cls._type(
            target=cls._arun_func,
            args=(func, state, *args),
            kwargs=kwargs,
        )
        task.start()
        while task.is_alive():
            await asleep(delay)
        task.join()
        return state.pop()

    @classmethod
    def new(
        cls,
        func: t.Callable[..., t.Any],
        /,
        *args: t.Any,
        probe_delay: t.Optional[t.PositiveRealNumber] = None,
        **kwargs: t.Any,
    ) -> t.Any:
        """
        Runs a sync function in another process or thread depending on
        the class calling this method, so that heavy cpu-bound
        computations, or blocking IO operations, can better coexist with
        asynchronous code.
        """
        delay = process_probe_delay(cls, probe_delay)
        state = cls._Manager().list()
        task = cls._type(
            target=cls._run_func,
            args=(func, state, *args),
            kwargs=kwargs,
        )
        task.start()
        while task.is_alive():
            sleep(delay)
        task.join()
        return state.pop()

    @staticmethod
    def _get_result(
        func: t.Callable[..., t.Any],
        /,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> t.Any:
        """
        Returns the result of `func` run in a new process / thread using
        the class' pool.
        """
        if is_async_function(func):
            run = new_event_loop().run_until_complete
            return run(func(*args, **kwargs))
        else:
            return func(*args, **kwargs)

    @staticmethod
    def _package_result_methods(
        future: t.Future, /, *, probe_delay: t.PositiveRealNumber
    ) -> t.Future:
        """
        Inserts methods in the `future` returned from a pool submission
        to provide a consistent interface for retrieving results.
        """

        async def aresult() -> t.Any:
            while not future.done():
                await asleep(probe_delay)
            return future._original_result()

        def result() -> t.Any:
            while not future.done():
                sleep(probe_delay)
            return future._original_result()

        future._original_result = future.result
        future.aresult = aresult
        future.result = result
        return future

    @classmethod
    async def asubmit(
        cls,
        func: t.Callable[..., t.Any],
        /,
        *args: t.Any,
        probe_delay: t.Optional[t.PositiveRealNumber] = None,
        **kwargs: t.Any,
    ) -> t.Future:
        """
        Submits an async, or synchronous `func` to a process pool or
        thread pool, depending on the class that calls this method, with
        the supplied `*args` & `**kwargs`, then returns the `Future`
        object that's created.
        """
        delay = process_probe_delay(cls, probe_delay)
        future = cls._pool.submit(cls._get_result, func, *args, **kwargs)
        return cls._package_result_methods(future, probe_delay=delay)

    @classmethod
    def submit(
        cls,
        func: t.Callable[..., t.Any],
        /,
        *args: t.Any,
        probe_delay: t.Optional[t.PositiveRealNumber] = None,
        **kwargs: t.Any,
    ) -> t.Future:
        """
        Submits a synchronous `func` to a process pool or thread pool,
        depending on the class that calls this method, with the supplied
        `*args` & `**kwargs`, then returns the `Future` object that's
        created.
        """
        delay = process_probe_delay(cls, probe_delay)
        future = cls._pool.submit(func, *args, **kwargs)
        return cls._package_result_methods(future, probe_delay=delay)

    @classmethod
    async def agather(
        cls,
        /,
        *functions: t.Callable[..., t.Any],
        args: t.Iterable[t.Any] = (),
        kwargs: t.Mapping[t.Hashable, t.Any] = {},
    ) -> t.List[t.Any]:
        """
        Sumbits all of the async or synchronous `functions` to the
        `Processes._pool` or `Threads._pool` with the given `args` &
        `kwargs`.
        """
        tasks = [
            await cls.asubmit(func, *args, **kwargs) for func in functions
        ]
        try:
            return [await task.aresult() for task in tasks]
        finally:
            for task in tasks:
                task.cancel()

    @classmethod
    def gather(
        cls,
        /,
        *functions: t.Callable[..., t.Any],
        args: t.Iterable[t.Any] = (),
        kwargs: t.Mapping[t.Hashable, t.Any] = {},
    ) -> t.List[t.Any]:
        """
        Sumbits all the `functions` to the `Processes._pool` or
        `Threads._pool` with the given `args` & `kwargs`.
        """
        tasks = [cls.submit(func, *args, **kwargs) for func in functions]
        try:
            return [task.result() for task in tasks]
        finally:
            for task in tasks:
                task.cancel()

    @classmethod
    def reset_pool(cls, /) -> None:
        """
        When a process or thread pool is broken by an abruptly exited,
        this method can be called to reset the class' pool object with
        a new instance.
        """
        cls._pool = cls._pool.__class__()


module_api = dict(
    ConcurrencyGuard=t.add_type(ConcurrencyGuard),
    ConcurrencyInterface=t.add_type(ConcurrencyInterface),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    process_probe_delay=process_probe_delay,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A type deriving database keys from passphrases using the passcrypt,
memory-hard hashing function.
"""

__all__ = ["AsyncProfileTokens", "ProfileTokens"]


from hashlib import sha3_512

from aiootp._typing import Typing as t
from aiootp._constants import MIN_KEY_BYTES
from aiootp._paths import AsyncSecurePath, SecurePath
from aiootp._paths import aread_salt_file, read_salt_file
from aiootp.asynchs import asleep
from aiootp.commons import FrozenSlots
from aiootp.generics import ahash_bytes, hash_bytes
from aiootp.keygens import Passcrypt

from .dbdomains import DBDomains


class AsyncProfileTokens(FrozenSlots):
    """
    Efficiently stores AsyncDatabase & Database profile token values
    which are used to more safely construct databases from potentially
    low entropy passphrases.
    """

    __slots__ = ("_gist", "_salt", "_salt_path", "login_key", "profile")

    def __init__(self) -> None:
        pass

    @classmethod
    async def _asummon_device_salt(cls, path: t.PathStr) -> bytes:
        """
        Generates a salt which is unique for each unique `path`
        directory that is given to this method. This is a static salt
        which provides an initial form of randomization to cryptographic
        material for all profiles saved under that directory.
        """
        salt_path = await AsyncSecurePath(
            path, key=DBDomains.DEVICE_SALT, _admin=True
        )
        return await aread_salt_file(salt_path)

    async def _asummon_profile_salt(
        self, path: t.PathStr
    ) -> t.Tuple[t.Path, bytes]:
        """
        Creates or loads a salt value saved on the user filesystem to
        help add more entropy to their key derivation functions when
        preparing to open a profile database.
        """
        salt_path = await AsyncSecurePath(path, key=self._gist)
        salt = await aread_salt_file(salt_path)
        return salt_path, salt

    async def _agenerate_profile_login_key(
        self,
        passphrase: bytes,
        **passcrypt_settings: t.Union[int, t.ConfigType],
    ) -> bytes:
        """
        Combines the output of the expensive key derivation functions &
        the salt stored on the filesystem gathered in preparation to
        safely open a profile database.
        """
        pcrypt = Passcrypt(tag_size=MIN_KEY_BYTES, **passcrypt_settings)
        return await pcrypt.anew(
            passphrase=await ahash_bytes(
                DBDomains.TMP_PREKEY,
                self._salt,
                self._gist,
                key=passphrase,
                hasher=sha3_512,
            ),
            salt=self._salt,
            aad=DBDomains.PROFILE_LOGIN_KEY,
        )

    async def agenerate(
        self,
        *credentials: bytes,
        username: bytes,
        passphrase: bytes,
        salt: bytes,
        aad: bytes,
        path: t.PathStr,
        **passcrypt_settings: t.Union[int, t.ConfigType],
    ) -> t.Self:
        """
        Runs a very expensive key derivation function to build keys
        for users to open a database with only access to potentially
        weakly entropic credentials & the filesystem.
        """
        device_salt = await self._asummon_device_salt(path=path)
        self._gist = await ahash_bytes(
            DBDomains.GIST,
            device_salt,
            salt,
            aad,
            username,
            *credentials,
            key=device_salt,
            hasher=sha3_512,
        )
        self._salt_path, self._salt = await self._asummon_profile_salt(
            path=path
        )
        self.login_key = await self._agenerate_profile_login_key(
            passphrase, **passcrypt_settings
        )
        return self

    async def acleanup(self) -> t.Self:
        """
        Removes cryptographic material from the object after use.
        """
        await asleep()
        object.__delattr__(self, "_gist")
        object.__delattr__(self, "_salt")
        object.__delattr__(self, "login_key")
        return self


class ProfileTokens(FrozenSlots):
    """
    Efficiently stores AsyncDatabase & Database profile token values
    which are used to more safely construct databases from potentially
    low entropy passphrases.
    """

    __slots__ = ("_gist", "_salt", "_salt_path", "login_key", "profile")

    def __init__(self) -> None:
        pass

    @classmethod
    def _summon_device_salt(cls, path: t.PathStr) -> bytes:
        """
        Generates a salt which is unique for each unique `path`
        directory that is given to this method. This is a static salt
        which provides an initial form of randomization to cryptographic
        material for all profiles saved under that directory.
        """
        salt_path = SecurePath(path, key=DBDomains.DEVICE_SALT, _admin=True)
        return read_salt_file(salt_path)

    def _summon_profile_salt(
        self, path: t.PathStr
    ) -> t.Tuple[t.Path, bytes]:
        """
        Creates or loads a salt value saved on the user filesystem to
        help add more entropy to their key derivation functions when
        preparing to open a profile database.
        """
        salt_path = SecurePath(path, key=self._gist)
        salt = read_salt_file(salt_path)
        return salt_path, salt

    def _generate_profile_login_key(
        self,
        passphrase: bytes,
        **passcrypt_settings: t.Union[int, t.ConfigType],
    ) -> bytes:
        """
        Combines the output of the expensive key derivation functions &
        the salt stored on the filesystem gathered in preparation to
        safely open a profile database.
        """
        pcrypt = Passcrypt(tag_size=MIN_KEY_BYTES, **passcrypt_settings)
        return pcrypt.new(
            passphrase=hash_bytes(
                DBDomains.TMP_PREKEY,
                self._salt,
                self._gist,
                key=passphrase,
                hasher=sha3_512,
            ),
            salt=self._salt,
            aad=DBDomains.PROFILE_LOGIN_KEY,
        )

    def generate(
        self,
        *credentials: bytes,
        username: bytes,
        passphrase: bytes,
        salt: bytes,
        aad: bytes,
        path: t.PathStr,
        **passcrypt_settings: t.Union[int, t.ConfigType],
    ) -> t.Self:
        """
        Runs a very expensive key derivation function to build keys
        for users to open a database with only access to potentially
        weakly entropic credentials & the filesystem.
        """
        device_salt = self._summon_device_salt(path=path)
        self._gist = hash_bytes(
            DBDomains.GIST,
            device_salt,
            salt,
            aad,
            username,
            *credentials,
            key=device_salt,
            hasher=sha3_512,
        )
        self._salt_path, self._salt = self._summon_profile_salt(path=path)
        self.login_key = self._generate_profile_login_key(
            passphrase, **passcrypt_settings
        )
        return self

    def cleanup(self) -> t.Self:
        """
        Removes cryptographic material from the object after use.
        """
        object.__delattr__(self, "_gist")
        object.__delattr__(self, "_salt")
        object.__delattr__(self, "login_key")
        return self


module_api = dict(
    AsyncProfileTokens=t.add_type(AsyncProfileTokens),
    ProfileTokens=t.add_type(ProfileTokens),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""\
a high-level async cryptographic anonymity library to scale, simplify, \
& automate privacy best practices for secure data & identity processing, \
communication, & storage.\
"""

__version__ = "0.23.12"


__license__ = "AGPLv3"


__author__ = "rmlibre@riseup.net"


__PUBLIC_ED25519_KEY__ = (
    "70d1740f2a439da98243c43a4d7ef1cf993b87a75f3bb0851ae79de675af5b3b"
)


__PUBLIC_X25519_KEY__ = (
    "4457276dbcae91cc5b69f1aed4384b9eb6f933343bb44d9ed8a80e2ce438a450"
)


import sys

from .commons import *
from .commons import remake_subpackage
from .asynchs import *
from .generics import *
from .randoms import *
from .ciphers import *
from .keygens import *
from .databases import *
from .__engagement.issue_reporting import report_security_issue


__all__ = [
    "commons",
    *commons.__all__,
    "asynchs",
    *asynchs.__all__,
    "generics",
    *generics.__all__,
    "randoms",
    *randoms.__all__,
    "ciphers",
    *ciphers.__all__,
    "keygens",
    *keygens.__all__,
    "databases",
    *databases.__all__,
]


subpackages = dict(
    _typing=_typing,
    _constants=_constants,
    _permutations=_permutations,
    commons=commons,
    asynchs=asynchs,
    generics=generics,
    randoms=randoms,
    ciphers=ciphers,
    keygens=keygens,
    databases=databases,
)


modules = dict(
    _exceptions=_exceptions,
    _debug_control=_debug_control,
    _gentools=_gentools,
    _paths=_paths,
)


module_api = dict(
    **{
        name: globals()[name]
        for name in __all__
        if not hasattr(globals()[name], "__spec__")
    },
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __author__=__author__,
    __loader__=__loader__,
    __license__=__license__,
    __package__=__package__,
    __version__=__version__,
    __PUBLIC_ED25519_KEY__=__PUBLIC_ED25519_KEY__,
    __PUBLIC_X25519_KEY__=__PUBLIC_X25519_KEY__,
    report_security_issue=report_security_issue,
)


aiootp = remake_subpackage(sys.modules[__name__])


del sys
del databases
del subpackages
del modules
del module_api
del __engagement
del remake_subpackage



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
The client side of a key exchange protocol which can provide
server-side identity authentication with ephemeral random secrets.
"""

__all__ = ["DoubleDiffieHellmanClient"]


from aiootp._typing import Typing as t
from aiootp._exceptions import Issue
from aiootp.commons import FrozenInstance
from aiootp.generics import Domains


class DoubleDiffieHellmanClient(FrozenInstance):
    """
    The client side of a key exchange protocol which can provide server-
    side identity authentication with ephemeral random secrets.

     _____________________________________
    |                                     |
    |          Protocol Diagram:          |
    |_____________________________________|

            -----------------          |         -----------------
            |  Client-side  |          |         |  Server-side  |
            -----------------          |         -----------------
                                       |
                                       |         X25519().generate() = key
                                       |
    client = X25519.dh2_client()       |           key.public_bytes = id_s
                                       |
    eph_c = client.send(id_s)       ------>
                                       |
                                       |         key.dh2_server() = server
                                       |
                                       |       server.receive(eph_c) = kdf
                                       |
                                    <------          server.send() = eph_s
                                       |
    kdf = client.receive(eph_s)        |
                                       |
    ------------------------------------------------------------------------
    """

    __slots__ = (
        "_kdf_type",
        "_key_exchange_type",
        "_my_ephemeral_key",
        "_peer_identity_key",
        "_peer_ephemeral_key",
        "_sanitize",
    )

    def __init__(
        self, /, key_exchange_type: type, *, kdf_type: type
    ) -> None:
        if not issubclass(kdf_type, t.DomainKDFType):
            raise Issue.value_must_be_subtype("KDF type", t.DomainKDFType)
        elif not issubclass(key_exchange_type, t.KeyExchangeType):
            raise Issue.value_must_be_subtype("KEX type", t.KeyExchangeType)
        self._kdf_type = kdf_type
        self._key_exchange_type = key_exchange_type
        self._sanitize = key_exchange_type._process_public_key

    async def asend(
        self, /, peer_identity_key: t.Union[bytes, t.PublicKeyType]
    ) -> bytes:
        """
        Receives the identity public key of the intended server & returns
        the instance's ephemeral public key to be sent to the server.
        """
        self._peer_identity_key = self._sanitize(peer_identity_key)
        self._my_ephemeral_key = await self._key_exchange_type().agenerate()
        return self._my_ephemeral_key.public_bytes

    def send(
        self, /, peer_identity_key: t.Union[bytes, t.PublicKeyType]
    ) -> bytes:
        """
        Receives the identity public key of the intended server & returns
        the instance's ephemeral public key to be sent to the server.
        """
        self._peer_identity_key = self._sanitize(peer_identity_key)
        self._my_ephemeral_key = self._key_exchange_type().generate()
        return self._my_ephemeral_key.public_bytes

    async def areceive(
        self, /, peer_ephemeral_key: t.Union[bytes, t.PublicKeyType]
    ) -> t.DomainKDFType:
        """
        Receives the ephemeral public key that could've been sent from a
        server, & returns the KDF object which has been primed with the
        public & secret values involed in the exchange.
        """
        my_ephemeral_key = self._my_ephemeral_key
        peer_identity_key = self._peer_identity_key
        peer_ephemeral_key = self._peer_ephemeral_key = self._sanitize(
            peer_ephemeral_key
        )
        shared_key_ad = await my_ephemeral_key.aexchange(peer_identity_key)
        shared_key_cd = await my_ephemeral_key.aexchange(peer_ephemeral_key)
        return self._kdf_type(
            Domains.DH2,
            my_ephemeral_key.public_bytes,
            self._peer_identity_key.public_bytes_raw(),
            self._peer_ephemeral_key.public_bytes_raw(),
            key=shared_key_ad + shared_key_cd,
        )

    def receive(
        self, /, peer_ephemeral_key: t.Union[bytes, t.PublicKeyType]
    ) -> t.DomainKDFType:
        """
        Receives the ephemeral public key that could've been sent from a
        server, & returns the KDF object which has been primed with the
        public & secret values involed in the exchange.
        """
        my_ephemeral_key = self._my_ephemeral_key
        peer_identity_key = self._peer_identity_key
        peer_ephemeral_key = self._peer_ephemeral_key = self._sanitize(
            peer_ephemeral_key
        )
        shared_key_ad = my_ephemeral_key.exchange(peer_identity_key)
        shared_key_cd = my_ephemeral_key.exchange(peer_ephemeral_key)
        return self._kdf_type(
            Domains.DH2,
            my_ephemeral_key.public_bytes,
            self._peer_identity_key.public_bytes_raw(),
            self._peer_ephemeral_key.public_bytes_raw(),
            key=shared_key_ad + shared_key_cd,
        )


module_api = dict(
    DoubleDiffieHellmanClient=t.add_type(DoubleDiffieHellmanClient),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A collection of (async) generator utilities.
"""

__all__ = []


import io
import asyncio
from types import GeneratorType
from types import AsyncGeneratorType
from collections.abc import Iterable, Iterator
from collections.abc import AsyncIterable, AsyncIterator
from inspect import isasyncgenfunction as is_async_gen_function
from inspect import isgeneratorfunction as is_generator_function

from ._typing import Typing as t
from ._constants import BIG
from ._exceptions import Issue, raise_exception


def is_async_iterable(obj: t.Any, /) -> bool:
    """
    Returns a bool of whether `obj` supports async iteration.
    """
    return isinstance(obj, AsyncIterable)


def is_iterable(obj: t.Any, /) -> bool:
    """
    Returns a bool of whether `obj` supports iteration.
    """
    return isinstance(obj, Iterable)


def is_async_iterator(obj: t.Any, /) -> bool:
    """
    Returns a bool of whether `obj` is an async iterator.
    """
    return isinstance(obj, AsyncIterator)


def is_iterator(obj: t.Any, /) -> bool:
    """
    Returns a bool of whether `obj` is an iterator.
    """
    return isinstance(obj, Iterator)


def is_async_generator(obj: t.Any, /) -> bool:
    """
    Returns a bool of whether `obj` is an async generator.
    """
    return isinstance(obj, AsyncGeneratorType)


def is_generator(obj: t.Any, /) -> bool:
    """
    Returns a bool of whether `obj` is an generator.
    """
    return isinstance(obj, GeneratorType)


async def abatch(
    sequence: bytes, /, *, size: int, buffer_type: type = io.BytesIO
) -> t.AsyncGenerator[bytes, None]:
    """
    Runs through a sequence & yields `size` sized chunks of the bytes
    sequence one chunk at a time. By default this async generator yields
    all elements in the sequence.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    sequence = 4 * b" Testing data..."

    async for piece in abatch(sequence, size=32):
        print(piece)
    >>> b' Testing data... Testing data...'
        b' Testing data... Testing data...'

    async for piece in abatch(sequence, size=64):
        print(piece)
    >>> b' Testing data... Testing data... Testing data... Testing data...'
    """
    try:
        read = buffer_type(sequence).read
        while True:
            await asyncio.sleep(0)
            yield read(size) or raise_exception(StopIteration)
    except StopIteration:
        pass


def batch(
    sequence: bytes, /, *, size: int, buffer_type: type = io.BytesIO
) -> t.Generator[bytes, None, None]:
    """
    Runs through a sequence & yields `size` sized chunks of the bytes
    sequence one chunk at a time. By default this generator yields all
    elements in the sequence.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    sequence = 4 * b" Testing data..."

    for piece in batch(sequence, size=32):
        print(piece)
    >>> b' Testing data... Testing data...'
        b' Testing data... Testing data...'

    for piece in batch(sequence, size=64):
        print(piece)
    >>> b' Testing data... Testing data... Testing data... Testing data...'
    """
    try:
        read = buffer_type(sequence).read
        while True:
            yield read(size) or raise_exception(StopIteration)
    except StopIteration:
        pass


async def aresize(
    sequence: t.Sequence[t.Any], /, size: int, *, blocks: int = 0
) -> t.AsyncGenerator[t.Sequence[t.Any], None]:
    """
    Runs through a `sequence` & yields `size` sized chunks of the
    sequence one chunk at a time. `blocks` is the total number of
    chunks allowed to be yielded from the generator. By default this
    generator yields all elements in the sequence.
    """
    length = len(sequence)
    if not blocks or (blocks * size >= length):
        blocks = length + size
    else:
        blocks = (blocks * size) + 1
    async for previous_end, end in azip(
        range(0, blocks, size), range(size, blocks, size)
    ):
        yield sequence[previous_end:end]


def resize(
    sequence: t.Sequence[t.Any], /, size: int, *, blocks: int = 0
) -> t.Generator[t.Sequence[t.Any], None, None]:
    """
    Runs through a `sequence` & yields `size` sized chunks of the
    sequence one chunk at a time. `blocks` is the total number of
    chunks allowed to be yielded from the generator. By default this
    generator yields all elements in the sequence.
    """
    length = len(sequence)
    if not blocks or (blocks * size >= length):
        blocks = length + size
    else:
        blocks = (blocks * size) + 1
    for previous_end, end in zip(
        range(0, blocks, size), range(size, blocks, size)
    ):
        yield sequence[previous_end:end]


async def aunpack(
    iterable: t.AsyncOrSyncIterable[t.Any], /
) -> t.AsyncGenerator[t.Any, None]:
    """
    Runs through an iterable &/or async iterable & yields elements one
    at a time.
    """
    if is_async_iterable(iterable):
        async for item in iterable:
            yield item
    else:
        for item in iterable:
            await asyncio.sleep(0)
            yield item


def unpack(
    iterable: t.Iterable[t.Any], /
) -> t.Generator[t.Any, None, None]:
    """
    Runs through an iterable & yields elements one at a time.
    """
    yield from iterable


async def azip(
    *iterables: t.AsyncOrSyncIterable[t.Any],
) -> t.AsyncGenerator[t.List[t.Any], None]:
    """
    Creates an asynchronous version of the `builtins.zip` function
    which is wrapped by the `Comprende` class.
    """
    coroutines = [aunpack(iterable).__anext__ for iterable in iterables]
    try:
        while True:
            yield [await coroutine() for coroutine in coroutines]
    except (StopAsyncIteration, StopIteration):
        pass


async def acycle(
    iterable: t.AsyncOrSyncIterable[t.Any], /
) -> t.AsyncGenerator[t.Any, None]:
    """
    Unendingly cycles in order over the elements of an async iterable.
    """
    results = []
    if is_async_iterable(iterable):
        async for result in iterable:
            yield result
            results.append(result)
    else:
        for result in iterable:
            await asyncio.sleep(0)
            yield result
            results.append(result)
    while results:
        for result in results:
            await asyncio.sleep(0)
            yield result


def cycle(iterable: t.Iterable[t.Any], /) -> t.Generator[t.Any, None, None]:
    """
    Unendingly cycles in order over the elements of a sync iterable.
    """
    results = []
    for result in iterable:
        yield result
        results.append(result)
    while results:
        for result in results:
            yield result


async def abytes_count(
    *, start: int = 0, size: int = 8, byte_order: str = BIG
) -> t.AsyncGenerator[bytes, None]:
    """
    Unendingly yields incrementing numbers starting from `start`.
    """
    index = start
    while True:
        await asyncio.sleep(0)
        yield index.to_bytes(size, byte_order)
        index += 1


def bytes_count(
    *, start: int = 0, size: int = 8, byte_order: str = BIG
) -> t.Generator[bytes, None, None]:
    """
    Unendingly yields incrementing numbers starting from `start`.
    """
    index = start
    while True:
        yield index.to_bytes(size, byte_order)
        index += 1


async def acount(*, start: int = 0) -> t.AsyncGenerator[int, None]:
    """
    Unendingly yields incrementing numbers starting from `start`.
    """
    if start.__class__ is not int:
        raise Issue.value_must_be_type("start", int)
    index = start
    while True:
        await asyncio.sleep(0)
        yield index
        index += 1


def count(*, start: int = 0) -> t.Generator[int, None, None]:
    """
    Unendingly yields incrementing numbers starting from `start`.
    """
    if start.__class__ is not int:
        raise Issue.value_must_be_type("start", int)
    index = start
    while True:
        yield index
        index += 1


async def abytes_range(
    *a: int, size: int = 8, byte_order: str = BIG
) -> t.AsyncGenerator[int, None]:
    """
    An async version of `builtins.range` wrapped by the `Comprende`
    class, & returns its values as bytes instead.
    """
    for result in range(*a):
        await asyncio.sleep(0)
        yield result.to_bytes(size, byte_order)


def bytes_range(
    *a: int, size: int = 8, byte_order: str = BIG
) -> t.Generator[int, None, None]:
    """
    A synchronous version of `builtins.range` which is wrapped by the
    `Comprende` class, & returns its values as bytes instead.
    """
    for result in range(*a):
        yield result.to_bytes(size, byte_order)


async def arange(*a: t.Optional[int]) -> t.AsyncGenerator[int, None]:
    """
    An async version of `builtins.range`.
    """
    for result in range(*a):
        await asyncio.sleep(0)
        yield result


async def acollate(
    *iterables: t.AsyncOrSyncIterable[t.Any],
) -> t.AsyncGenerator[t.Any, None]:
    """
    Takes a collection of iterables &/or async iterables & exhausts them
    one at a time from left to right.
    """
    for iterable in iterables:
        if is_async_iterable(iterable):
            async for result in iterable:
                yield result
        else:
            for result in iterable:
                await asyncio.sleep(0)
                yield result


def collate(
    *iterables: t.Iterable[t.Any],
) -> t.Generator[t.Any, None, None]:
    """
    Takes a collection of iterables & exhausts them one at a time from
    left to right.
    """
    for iterable in iterables:
        yield from iterable


async def apopleft(
    queue: t.SupportsPopleft, /
) -> t.AsyncGenerator[t.Any, None]:
    """
    An async generator which calls the `popleft()` method on `queue`
    for every iteration, & exits on `IndexError`.
    """
    try:
        while True:
            await asyncio.sleep(0)
            yield queue.popleft()
    except IndexError:
        pass


def popleft(queue: t.SupportsPopleft, /) -> t.Generator[t.Any, None, None]:
    """
    A generator which calls the `popleft()` method on `queue` for
    every iteration, & exits on `IndexError`.
    """
    try:
        while True:
            yield queue.popleft()
    except IndexError:
        pass


module_api = dict(
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    abatch=abatch,
    abytes_count=abytes_count,
    abytes_range=abytes_range,
    acollate=acollate,
    acount=acount,
    acycle=acycle,
    apopleft=apopleft,
    arange=arange,
    aresize=aresize,
    aunpack=aunpack,
    azip=azip,
    batch=batch,
    bytes_count=bytes_count,
    bytes_range=bytes_range,
    collate=collate,
    count=count,
    cycle=cycle,
    is_async_iterable=is_async_iterable,
    is_async_iterator=is_async_iterator,
    is_async_generator=is_async_generator,
    is_async_gen_function=is_async_gen_function,
    is_generator=is_generator,
    is_generator_function=is_generator_function,
    is_iterable=is_iterable,
    is_iterator=is_iterator,
    popleft=popleft,
    resize=resize,
    unpack=unpack,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *


# NOTE: The tests run in alphabetical order, so a Z is prepended to their
#       names. This way time will naturally pass from the start of the
#       tests to the end, eliminating idle waiting time.


class TestZCipherTimeToLive:
    async def test_async_json_decrypt(self) -> None:
        problem = (  # fmt: skip
            "Life-time for async json ciphertext is malfunctioning."
        )
        with Ignore(
            TimestampExpired, if_else=violation(problem)
        ) as ignored:
            relay = ignored
            await ttl_test_cipher.cipher.ajson_decrypt(
                atest_json_ciphertext, aad=ttl_test_cipher.aad, ttl=1
            )
        assert relay.error.expired_by >= 1

    async def test_sync_json_decrypt(self) -> None:
        problem = (  # fmt: skip
            "Life-time for sync json ciphertext is malfunctioning."
        )
        with Ignore(
            TimestampExpired, if_else=violation(problem)
        ) as ignored:
            relay = ignored
            ttl_test_cipher.cipher.json_decrypt(
                test_json_ciphertext, aad=ttl_test_cipher.aad, ttl=1
            )
        assert relay.error.expired_by >= 1

    async def test_async_read_token(self) -> None:
        problem = (  # fmt: skip
            "Life-time for async tokens is malfunctioning."
        )
        with Ignore(
            TimestampExpired, if_else=violation(problem)
        ) as ignored:
            relay = ignored
            await ttl_test_cipher.cipher.aread_token(
                atest_token_ciphertext, aad=ttl_test_cipher.aad, ttl=1
            )
        assert relay.error.expired_by >= 1

    async def test_sync_read_token(self) -> None:
        problem = (  # fmt: skip
            "Life-time for sync tokens is malfunctioning."
        )
        with Ignore(
            TimestampExpired, if_else=violation(problem)
        ) as ignored:
            relay = ignored
            ttl_test_cipher.cipher.read_token(
                test_token_ciphertext, aad=ttl_test_cipher.aad, ttl=1
            )
        assert relay.error.expired_by >= 1

    async def test_async_cipher_stream(self) -> None:
        problem = (  # fmt: skip
            "Life-time for async cipher streams is malfunctioning."
        )
        enc_stream, ciphertext = attl_cipher_stream, attl_stream_ciphertext
        with Ignore(
            TimestampExpired, if_else=violation(problem)
        ) as ignored:
            relay = ignored
            dec_stream = await ttl_test_cipher.cipher.astream_decrypt(
                salt=enc_stream.salt,
                aad=enc_stream.aad,
                iv=enc_stream.iv,
                ttl=1,
            )
            await dec_stream.abuffer(ciphertext)
            b"".join([block async for block in dec_stream])
        assert relay.error.expired_by >= 1

    async def test_sync_cipher_stream(self) -> None:
        problem = (  # fmt: skip
            "Life-time for sync cipher streams is malfunctioning."
        )
        enc_stream, ciphertext = ttl_cipher_stream, ttl_stream_ciphertext
        with Ignore(
            TimestampExpired, if_else=violation(problem)
        ) as ignored:
            relay = ignored
            dec_stream = ttl_test_cipher.cipher.stream_decrypt(
                salt=enc_stream.salt,
                aad=enc_stream.aad,
                iv=enc_stream.iv,
                ttl=1,
            )
            b"".join(dec_stream.buffer(ciphertext))
        assert relay.error.expired_by >= 1


class TestZPasscryptTimeToLive:
    async def test_async_verfiy_ttl(self) -> None:
        problem = (  # fmt: skip
            "Life-time for async passcrypt hashes are malfunctioning."
        )
        with Ignore(
            TimestampExpired, if_else=violation(problem)
        ) as ignored:
            relay = ignored
            await Passcrypt.averify(
                aexpired_passcrypt_hash, passphrase_0, ttl=1
            )
        assert relay.error.expired_by >= 1

    async def test_sync_verfiy_ttl(self) -> None:
        problem = (  # fmt: skip
            "Life-time for sync passcrypt hashes are malfunctioning."
        )
        with Ignore(
            TimestampExpired, if_else=violation(problem)
        ) as ignored:
            relay = ignored
            Passcrypt.verify(expired_passcrypt_hash, passphrase_0, ttl=1)
        assert relay.error.expired_by >= 1


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


import io
from collections import deque

from conftest import *

from aiootp import _gentools as gentools


class TestInspectTools:
    async def test_is_async_iterable(self) -> None:
        async_iterable = gentools.aunpack("test")
        iterable = gentools.unpack("test")

        assert gentools.is_async_iterable(async_iterable)
        assert not gentools.is_async_iterable(iterable)

    async def test_is_async_iterator(self) -> None:
        async_iterator = gentools.aunpack("test").__aiter__()
        iterator = gentools.unpack("test").__iter__()

        assert gentools.is_async_iterator(async_iterator)
        assert not gentools.is_async_iterator(iterator)

    async def test_is_async_generator(self) -> None:
        async_generator = gentools.aunpack("test")
        generator = gentools.unpack("test")

        assert gentools.is_async_generator(async_generator)
        assert not gentools.is_async_generator(generator)

    async def test_is_async_gen_function(self) -> None:
        async_generator_function = gentools.aunpack
        generator_function = gentools.unpack

        assert gentools.is_async_gen_function(async_generator_function)
        assert not gentools.is_async_gen_function(generator_function)

    async def test_is_generator(self) -> None:
        async_generator = gentools.aunpack("test")
        generator = gentools.unpack("test")

        assert not gentools.is_generator(async_generator)
        assert gentools.is_generator(generator)

    async def test_is_generator_function(self) -> None:
        async_generator_function = gentools.aunpack
        generator_function = gentools.unpack

        assert not gentools.is_generator_function(async_generator_function)
        assert gentools.is_generator_function(generator_function)

    async def test_is_iterable(self) -> None:
        async_iterable = gentools.aunpack("test")
        iterable = gentools.unpack("test")

        assert not gentools.is_iterable(async_iterable)
        assert gentools.is_iterable(iterable)

    async def test_is_iterator(self) -> None:
        async_iterator = gentools.aunpack("test").__aiter__()
        iterator = gentools.unpack("test").__iter__()

        assert not gentools.is_iterator(async_iterator)
        assert gentools.is_iterator(iterator)


class TestGenerators:
    async def test_bytes_count(self) -> None:
        for start in (0, 33, 256, 1024):
            for size in (4, 8):
                for order in ("little", "big"):
                    count = start
                    async for index, aindex in gentools.azip(
                        gentools.bytes_count(
                            start=start, size=size, byte_order=order
                        ),
                        gentools.abytes_count(
                            start=start, size=size, byte_order=order
                        ),
                    ):
                        assert index == aindex
                        assert index == count.to_bytes(size, order)

                        count += 1
                        if count > start + 4:
                            break

    async def test_counts(self) -> None:
        for start in (0, 33, 256, 1024):
            count = start
            async for index, aindex in gentools.azip(
                gentools.count(start=start),
                gentools.acount(start=start),
            ):
                assert index == aindex
                assert index == count

                count += 1
                if count > start + 4:
                    break

    async def test_start_must_be_int_for_counts(self) -> None:
        problem = "A non-int `start` value was allowed."
        for start in (0.0, None, "test", b"test"):
            with Ignore(TypeError, if_else=violation(problem)):
                async for _ in gentools.acount(start=start):
                    pass
            with Ignore(TypeError, if_else=violation(problem)):
                for _ in gentools.count(start=start):
                    pass

    async def test_bytes_ranges(self) -> None:
        for start in (0, 33, 256):
            end = start + 4
            for size in (4, 8):
                for order in ("little", "big"):
                    for skip in (1, 2):
                        count = start
                        async for index, aindex in gentools.azip(
                            gentools.bytes_range(
                                start,
                                end,
                                skip,
                                size=size,
                                byte_order=order,
                            ),
                            gentools.abytes_range(
                                start,
                                end,
                                skip,
                                size=size,
                                byte_order=order,
                            ),
                        ):
                            assert index == aindex
                            assert index == count.to_bytes(size, order)
                            assert start <= count < end
                            count += skip

    async def test_ranges(self) -> None:
        for start in (0, 33, 256):
            end = start + 4
            for skip in (1, 2):
                count = start
                async for index, aindex in gentools.azip(
                    range(start, end, skip),
                    gentools.arange(start, end, skip),
                ):
                    assert index == aindex
                    assert index == count
                    assert start <= count < end
                    count += skip

    async def test_batches(self) -> None:
        for blocks in (1, 2, 3):
            for size in (4, 8, 13):
                for unit, cls in ((b"a", io.BytesIO), ("a", io.StringIO)):
                    result = unit.__class__()
                    data = blocks * size * unit
                    async for block, ablock in gentools.azip(
                        gentools.batch(data, size=size, buffer_type=cls),
                        gentools.abatch(data, size=size, buffer_type=cls),
                    ):
                        assert block == ablock
                        assert len(block) == size
                        result += block
                    assert result == data

    async def test_resizes(self) -> None:
        for blocks in (1, 2, 3):
            for size in (4, 6, 8):
                for unit, cls in (
                    ([None], list),
                    ((None,), tuple),
                    ("a", str),
                ):
                    result = cls()
                    data = 3 * size * unit
                    async for block, ablock in gentools.azip(
                        gentools.resize(data, size=size, blocks=blocks),
                        gentools.aresize(data, size=size, blocks=blocks),
                    ):
                        assert block == ablock
                        assert len(block) == size
                        result = result + block
                    assert result == data[: blocks * size]

    async def test_poplefts(self) -> None:
        size = 32
        count = 0
        container = deque(range(size))
        acontainer = deque(range(size))
        for ordering in (
            (gentools.popleft(container), gentools.apopleft(acontainer)),
            (gentools.apopleft(acontainer), gentools.popleft(container)),
        ):
            async for item, aitem in gentools.azip(*ordering):
                assert item == aitem
                assert item == count
                count += 1
                assert len(container) == len(acontainer)
                assert len(container) == size - count

    async def test_unpacks(self) -> None:
        count = 0
        size = 32
        iterable = tuple(range(size))
        aiterable = gentools.arange(size)
        async for index, aindex, aaindex in gentools.azip(
            gentools.unpack(iterable),
            gentools.aunpack(iterable),
            gentools.aunpack(aiterable),
        ):
            assert index == aindex
            assert index == aaindex
            assert index == count
            count += 1

    async def test_cycles(self) -> None:
        count = 0
        size = 16
        iteration = 0
        iterable = tuple(range(size))
        aiterable = gentools.arange(size)
        async for index, aindex, aaindex in gentools.azip(
            gentools.cycle(iterable),
            gentools.acycle(iterable),
            gentools.acycle(aiterable),
        ):
            assert index == aindex
            assert index == aaindex
            assert index == count
            count = (count + 1) % size
            if count == 0:
                iteration += 1
            if iteration > 2:
                break

    async def test_collates(self) -> None:
        count = 0
        arange = gentools.arange
        iterable = ((0, 1, 2), (3, 4, 5), (6, 7, 8))
        aiterable = (arange(0, 3), (3, 4, 5), arange(6, 9))
        async for index, aindex in gentools.azip(
            gentools.collate(*iterable),
            gentools.acollate(*aiterable),
        ):
            assert index == aindex
            assert index == count
            count += 1


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A definition of shared properties between cipher streaming types.
"""

__all__ = ["AuthFail", "CipherStreamProperties"]


from aiootp._typing import Typing as t
from aiootp._exceptions import TimestampExpired, IncoherentConcurrencyState
from aiootp._exceptions import InvalidBlockID, InvalidSHMAC
from aiootp.asynchs import asleep
from aiootp.commons import OpenFrozenSlots
from aiootp.generics import abytes_are_equal, bytes_are_equal


class AuthFail(OpenFrozenSlots):
    """
    Creates efficient containers for data lost in a buffer during
    authentication failure of a block ID in (Async)DecipherStream
    objects.
    """

    __slots__ = ("block_id", "block", "buffer")

    def __init__(
        self, block_id: bytes, block: bytes, buffer: t.Callable
    ) -> None:
        self.block_id = block_id
        self.block = block
        self.buffer = buffer


class CipherStreamProperties:
    """
    A definition of shared properties between cipher streaming types.
    """

    IncoherentConcurrencyState: type = IncoherentConcurrencyState
    InvalidBlockID: type = InvalidBlockID
    InvalidSHMAC: type = InvalidSHMAC
    TimestampExpired: type = TimestampExpired

    @property
    def PACKETSIZE(self) -> int:
        """
        Returns the number of combined bytes each iteration of the
        stream will produce. Equal to the header bytes + the blocksize.
        """
        return self._config.PACKETSIZE

    @property
    def aad(self) -> bytes:
        """
        An arbitrary bytes value that a user decides to categorize
        keystreams. It's authenticated as associated data & safely
        differentiates keystreams as a tweak when it's unique for
        each permutation of `key`, `salt`, & `iv`.
        """
        return self._key_bundle.aad

    @property
    def salt(self) -> bytes:
        """
        A [pseudo]random salt that may be supplied by the user. By
        default it's sent in the clear attached to the ciphertext.
        Thus it may simplify implementing efficient features, such
        as search or routing, though care must still be taken when
        considering how leaking such metadata may be harmful.

        Keeping this value constant is strongly discouraged. Though,
        the cipher's salt misuse-reuse resistance is ruled by the
        combination of the automatically incorporated `timestamp`,
        `iv`, & `siv_key`. The risk calculation starts with setting
        r = len(iv + siv_key) / 3. Then, all else staying constant,
        once 256**r messages are encrypted within a second, each
        additional encrypted message within that same second begins
        to have more than a 256**(-r) chance of generating a repeat
        context.

        See: https://github.com/rmlibre/aiootp/issues/16
        """
        return self._key_bundle.salt

    @property
    def iv(self) -> bytes:
        """
        An ephemeral, uniform, random value that's generated by
        the encryption algorithm.
        """
        return self._key_bundle.iv

    async def _aconstant_time_final_context_is_done(self) -> bool:
        """
        Tests in constant time whether the queue token set during the
        (a)finalize call has already been popped from the digesting queue.
        If so, the stream has been closed so returns `True`. Otherwise
        returns `False`.
        """
        await asleep()
        if not self._finalizing_now:
            return False

        final_token = self._finalizing_now[0]
        membership_tests = [
            await abytes_are_equal(queued_token, final_token)
            for queued_token in self._digesting_now
        ]
        return not sum(membership_tests)

    def _constant_time_final_context_is_done(self) -> bool:
        """
        Tests in constant time whether the queue token set during the
        (a)finalize call has already been popped from the digesting queue.
        If so, the stream has been closed so returns `True`. Otherwise
        returns `False`.
        """
        if not self._finalizing_now:
            return False

        final_token = self._finalizing_now[0]
        membership_tests = (
            bytes_are_equal(queued_token, final_token)
            for queued_token in self._digesting_now
        )
        return not sum(membership_tests)


module_api = dict(
    AuthFail=t.add_type(AuthFail),
    CipherStreamProperties=t.add_type(CipherStreamProperties),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Classes using a `slots_types` mapping to enforce variable-type
associations on the variables named in their `__slots__`.
"""

__all__ = [
    "FrozenTypedSlots",
    "OpenFrozenTypedSlots",
    "OpenTypedSlots",
    "TypedSlots",
]


from aiootp._typing import Typing as t
from aiootp._exceptions import Issue, TypeUncheckableAtRuntime
from aiootp._exceptions import MissingDeclaredVariables

from .slots import Slots, OpenFrozenSlots


class TypedSlots(Slots):
    """
    Allows subclasses to define a `slots_types` mapping that enforces
    variable-type associations on the variables named in `__slots__`."

    Masked repr.
    Mutable instance.

     _____________________________________
    |                                     |
    |   Stability of Assignment Styles:   |
    |_____________________________________|


    from aiootp.commons.typed_slots import TypedSlots

    class HybridNamespace(TypedSlots):
        __slots__ = ("attr", "__dict__")
        slots_types = dict(attr=str)

    hybrid = HybridNamespace()

    ✔ hybrid.attr = "value"                # supported
    ✔ hybrid["attr"] = "value"             # supported
    ✔ setattr(hybrid, "attr", "value")     # supported

    ❌ hybrid.__dict__["attr"] = "value"    # unsupported


    # See: https://github.com/rmlibre/aiootp/pull/11
    """

    __slots__ = ()

    slots_types: t.Mapping[str, t.Any] = dict()

    @classmethod
    def _copy_acceptable_type_declarations(
        cls, /, base: type, slots_types: t.Mapping[str, t.Any]
    ) -> None:
        """
        Traverses a `base` class' slots types declarations & tests them
        for validity & consistency. Copies declarations into `slots_types`
        for the subclass to install as their own.

        Raises `TypeUncheckableAtRuntime` if a type is invalid as a
        run-time type checker.

        Raises `MissingDeclaredVariables` if a type is declared in the
        `base`'s `slots_types` but not in its (or one of its superclass')
        `__slots__`.
        """
        if not issubclass(base, TypedSlots):
            return
        for name, value in base.slots_types.items():
            try:
                isinstance(value, value)
                issubclass(type, value)
            except TypeError as error:
                raise TypeUncheckableAtRuntime(name, value) from error
            if name not in cls.__slots__:
                raise MissingDeclaredVariables(
                    name, found_in="slots_types", missed="__slots__"
                )
            elif name != "__dict__":
                slots_types[name] = value

    @classmethod
    def _make_frozen_class_slots_types_container(
        cls, slots_types: t.Mapping[str, type], /
    ) -> OpenFrozenSlots:
        """
        Creates a class-specific type to govern type correctness.
        """
        cls_name = f"{cls.__qualname__}SlotsTypes"
        cls_dict = dict(
            __slots__=tuple(slots_types),
            __module__=__name__,
        )
        container = type(cls_name, (OpenFrozenSlots,), cls_dict)
        return container(**slots_types)

    @classmethod
    def _make_frozen_class_slots_types(cls, /) -> OpenFrozenSlots:
        """
        Creates a class-specific type to govern type correctness by
        traversing the method resolution order, then collecting the
        valid slots types declarations.

        Raises `MissingDeclaredVariables` if a type is declared in the
        `__slots__` of a class in the mro but not in its `slots_types`.
        """
        slots_types = {}
        for base in reversed(cls.__mro__):
            cls._copy_acceptable_type_declarations(base, slots_types)
        diff = (
            set(cls.__slots__)
            .difference({"__dict__"})
            .symmetric_difference(slots_types)
        )
        if diff:
            raise MissingDeclaredVariables(
                *diff, found_in="__slots__", missed="slots_types"
            )
        return cls._make_frozen_class_slots_types_container(slots_types)

    def __init_subclass__(cls, /, *a: t.Any, **kw: t.Any) -> None:
        """
        Installs a prepared an class-specific type to govern type
        correctness to all subclasses.
        """
        super().__init_subclass__(*a, **kw)
        cls.slots_types = cls._make_frozen_class_slots_types()

    def _validate_type(self, name: str, value: t.Any, /) -> None:
        """
        Validates the type of the `value` based on the class' type
        definition of the `name` attribute.
        """
        if name not in self.slots_types:
            return

        value_type = self.slots_types[name]
        if not isinstance(value, value_type):
            raise Issue.value_must_be_type(name, value_type)

    def __setitem__(self, name: str, value: t.Any, /) -> None:
        """
        Transforms bracket lookup into dotted access on the instance.
        Sets the attribute if the type is correct.
        """
        self._validate_type(name, value)
        super().__setitem__(name, value)


class OpenTypedSlots(TypedSlots):
    """
    Allows subclasses to define a `slots_types` mapping that enforces
    variable-type associations on the variables named in `__slots__`."

    Unmasked repr.
    Mutable instance.
    """

    __slots__ = ()

    slots_types: t.Mapping[str, t.Any] = dict()

    def __repr__(self, /, *, mask: bool = False) -> str:
        """
        Pretty displays the instance & its attributes.
        """
        return super().__repr__(mask=mask)


class FrozenTypedSlots(TypedSlots):
    """
    Allows subclasses to define a `slots_types` mapping that enforces
    variable-type associations on the variables named in `__slots__`."

    Masked repr.
    Immutable instance. (set once)
    """

    __slots__ = ()

    slots_types: t.Mapping[str, t.Any] = dict()

    def __setitem__(self, name: str, value: t.Any, /) -> None:
        """
        Denies setting attributes after they have already been set.
        """
        if name in self:
            raise Issue.cant_reassign_attribute(name)

        super().__setitem__(name, value)

    def __delitem__(self, name: str, /) -> None:
        """
        Denies deleting attributes.
        """
        raise Issue.cant_deassign_attribute(name)


class OpenFrozenTypedSlots(FrozenTypedSlots):
    """
    Allows subclasses to define a `slots_types` mapping that enforces
    variable-type associations on the variables named in `__slots__`."

    Unmasked repr.
    Immutable instance. (set once)
    """

    __slots__ = ()

    slots_types: t.Mapping[str, t.Any] = dict()

    def __repr__(self, /, *, mask: bool = False) -> str:
        """
        Pretty displays the instance & its attributes.
        """
        return super().__repr__(mask=mask)


module_api = dict(
    FrozenTypedSlots=t.add_type(FrozenTypedSlots),
    OpenFrozenTypedSlots=t.add_type(OpenFrozenTypedSlots),
    OpenTypedSlots=t.add_type(OpenTypedSlots),
    TypedSlots=t.add_type(TypedSlots),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A configuration type for `AffinePermutation`.
"""

__all__ = ["AffinePermutationConfig"]


from math import log2

from aiootp._typing import Typing as t
from aiootp._constants import datasets, BIG
from aiootp._exceptions import Issue
from aiootp._gentools import count as counter
from aiootp.commons import Config
from aiootp.generics import Domains


class AffinePermutationConfig(Config):
    """
    Configuration type for `AffinePermutation`.
    """

    __slots__ = (
        "AAD",
        "SIZE",
        "KEY_SIZE",
        "MAX",
        "PRIME",
        "MULTIPLICATIVE_KEY",
        "INVERSE_KEY",
    )

    slots_types: t.Mapping[str, type] = dict(
        AAD=bytes,
        SIZE=int,
        KEY_SIZE=int,
        MAX=int,
        PRIME=int,
        MULTIPLICATIVE_KEY=int,
        INVERSE_KEY=int,
    )

    def _process_size(self, size: int) -> int:
        """
        Returns `size` only if it passes non-type based validity checks,
        otherwise raises `ValueError`.
        """
        if 4096 >= size > 0:
            return size
        else:
            raise Issue.value_must("size", "be > 0 and <= 4096")

    @staticmethod
    def _make_multiplier_mask(bit_length: int) -> int:
        """
        Returns an integer mask to remove half of the most significant
        bits of a multiplier. This will reduce the side-channel leakage
        of the product when it's combined with the addition key before
        modular reduction.
        """
        top_bit = 1 << (bit_length // 2)
        mask = top_bit - 1
        return top_bit, mask

    @staticmethod
    def _count_bit_switches(number: int) -> int:
        """
        Allows determining if the provided number has a sufficiently
        distributed number of set bits.
        """
        count = 0
        binary = bin(number)[2:]
        current = binary[:1]
        for bit in binary:
            if bit != current:
                count += 1
                current = bit
        return count

    @classmethod
    def is_likely_safe_multiplier(cls, multiplier: int, prime: int) -> bool:
        """
        Returns `True` if a `multiplier` key is odd, about ~half the bit-
        length of the `prime`, & has a bit pattern with about an equal
        number of 1s & 0s with some clumping. Otherwise returns `False`.
        """
        if prime < 257:
            raise Issue.value_must("prime", "be > 256")
        prime_size = prime.bit_length()
        sqrt_prime = 1 << (prime_size // 2)
        bit_flips = cls._count_bit_switches(multiplier)
        span = int(log2(prime_size) - 1)
        max_flips, min_flips = bit_flips + span, bit_flips - span
        return bool(
            (multiplier & 1)
            and ((sqrt_prime << 2) > multiplier > sqrt_prime)
            and (max_flips > (prime_size / 4) > min_flips)
        )

    def _derive_new_multiplier(self, prime: int) -> int:
        """
        Derives the static multiplicative key for a provided `prime`
        number, the size of the instance permutation, & its associated
        data. This amortizes the relatively expensive derivation costs
        of finding the multiplicative key, & its inverse.
        """

        def encode(i: int) -> bytes:
            return Domains.encode_constant(
                f"multiplicative_key_{i}_{prime}",
                domain=b"affine_permutation_constant",
                aad=self.AAD,
                size=self.SIZE,
            )

        top_bit, mask = self._make_multiplier_mask(prime.bit_length())
        for i in counter():
            key = top_bit | (int.from_bytes(encode(i), BIG) & mask)
            if self.is_likely_safe_multiplier(key, prime):
                return key

    def __init__(self, *, size: int, aad: bytes = b"") -> None:
        """
        Each instance generates the dynamic values determined by its
        specific `SIZE` & `AAD`.
        """
        self.AAD = aad
        self.SIZE = self._process_size(size)
        self.KEY_SIZE = 2 * self.SIZE
        self.MAX = (1 << (8 * self.SIZE)) - 1
        self.PRIME = datasets.PRIMES[8 * self.SIZE + 1][0]
        self.MULTIPLICATIVE_KEY = self._derive_new_multiplier(self.PRIME)
        self.INVERSE_KEY = pow(self.MULTIPLICATIVE_KEY, -1, self.PRIME)


module_api = dict(
    AffinePermutationConfig=t.add_type(AffinePermutationConfig),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Tuples defining class categories for runtime type enforcement.
"""

__all__ = ["BytesLike"]


BytesLike = (bytes, bytearray)


module_api = dict(
    BytesLike=BytesLike,
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


import io
from collections import deque

from conftest import *

from aiootp._constants.misc import BIG
from aiootp._constants.datasets import Tables
from aiootp.generics.transform import abase_as_int, base_as_int
from aiootp.generics.transform import aint_as_base, int_as_base


class TestBaseConversions:
    async def test_base_as_int_correctness(self) -> None:
        for string, answer, table in (
            ("0", 0, Tables.HEX),
            ("a31b", 41755, Tables.HEX),
            ("ff", 255, Tables.HEX),
            (b"\x00", 0, Tables.BYTES),
            (b"\xa3\x1b", 41755, Tables.BYTES),
            (b"\xff", 255, Tables.BYTES),
        ):
            assert answer == await abase_as_int(string, table=table)
            assert answer == base_as_int(string, table=table)

        problem = (  # fmt: skip
            "A character not a part of the base's table didn't error."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            await abase_as_int("g", table=Tables.HEX)
        with Ignore(ValueError, if_else=violation(problem)):
            base_as_int("g", table=Tables.HEX)

    async def test_int_as_base_correctness(self) -> None:
        for number, answer, table in (
            (0, "0", Tables.HEX),
            (41755, "a31b", Tables.HEX),
            (255, "ff", Tables.HEX),
            (0, b"\x00", Tables.BYTES),
            (41755, b"\xa3\x1b", Tables.BYTES),
            (255, b"\xff", Tables.BYTES),
        ):
            assert answer == await aint_as_base(number, table=table)
            assert answer == int_as_base(number, table=table)


class TestDomains:
    async def test_encoding_methods(self) -> None:
        # fmt: off
        for constant in (b"", "string constant...", b"bytes constant..."):
            for aad in (b"", b"associated data..."):
                for domain in (b"", b"domain..."):
                    assert (
                        await Domains.aencode_constant(constant, aad=aad, domain=domain)
                        ==  Domains.encode_constant(constant, aad=aad, domain=domain)
                    )
        # fmt: on


class TestEncodingUtilities:
    async def test_key_must_not_be_falsey(self) -> None:
        problem = (  # fmt: skip
            "A falsey key was allowed."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            await generics.canon.aencode_key(key=b"", blocksize=32)
        with Ignore(ValueError, if_else=violation(problem)):
            generics.canon.encode_key(key=b"", blocksize=32)

    async def test_item_size_mismatches_caught_during_decoding(
        self,
    ) -> None:
        # fmt: off
        problem = (
            "A mismatch between item length & declared length was allowed."
        )
        int_bytes = 1
        item = b"test"
        item_length = len(item).to_bytes(int_bytes, BIG)
        single_encoded_item = item_length + item[:-1]
        with Ignore(ValueError, if_else=violation(problem)):
            [item async for item in generics.canon.adecode_items(
                read=io.BytesIO(single_encoded_item).read,
                item_count=1,
                int_bytes=1,
            )]
        with Ignore(ValueError, if_else=violation(problem)):
            list(generics.canon.decode_items(
                read=io.BytesIO(single_encoded_item).read,
                item_count=1,
                int_bytes=1,
            ))
        # fmt: on

    async def test_missing_padding_metadata_throws_error(self) -> None:
        problem = (  # fmt: skip
            "Padding test ran without the necessary metadata."
        )
        blocksize = (32).to_bytes(1, BIG)
        with Ignore(ValueError, if_else=violation(problem)):
            generics.canon.test_canonical_padding(
                read=io.BytesIO().read,
                items=deque([blocksize]),  # <--- woops, missing: pad
                total_size=64,
            )

    async def test_non_blocksize_multiple_declared_length_throws_error(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "Padding test ran with a length declaration not equal to a "
            "multiple of the blocksize."
        )
        blocksize = (32).to_bytes(1, BIG)
        pad = b"\x00"
        with Ignore(ValueError, if_else=violation(problem)):
            generics.canon.test_canonical_padding(
                read=io.BytesIO().read,
                items=deque([blocksize, pad]),
                total_size=67,
            )

    async def test_declared_pad_value_must_be_all_that_remains(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "Padding test ran with an invalid padding value."
        )
        blocksize = (32).to_bytes(1, BIG)
        pad = b"\x00"
        generics.canon.test_canonical_padding(
            read=io.BytesIO(32 * pad).read,
            items=deque([blocksize, pad]),
            total_size=64,
        )
        with Ignore(ValueError, if_else=violation(problem)):
            generics.canon.test_canonical_padding(
                read=io.BytesIO(32 * b"A").read,
                items=deque([blocksize, pad]),
                total_size=64,
            )


class TestCanonicalPack:
    async def test_empty_pad_is_invalid(self) -> None:
        problem = (  # fmt: skip
            "An empty padding value didn't raise a `TypeError`."
        )
        with Ignore(TypeError, if_else=violation(problem)):
            canonical_pack(b"test", pad=b"")
        async with Ignore(TypeError, if_else=violation(problem)):
            await acanonical_pack(b"test", pad=b"")

    async def test_zero_blocksize_is_invalid(self) -> None:
        problem = (  # fmt: skip
            "A negative blocksize value didn't raise a `OverflowError`."
        )
        with Ignore(ZeroDivisionError, if_else=violation(problem)):
            canonical_pack(b"test", blocksize=0)
        async with Ignore(ZeroDivisionError, if_else=violation(problem)):
            await acanonical_pack(b"test", blocksize=0)

    async def test_negative_integer_blocksize_is_invalid(self) -> None:
        problem = (  # fmt: skip
            "A negative blocksize value didn't raise a `OverflowError`."
        )
        test_values = (-168, -136, -2, -1)
        for blocksize in test_values:
            with Ignore(OverflowError, if_else=violation(problem)):
                canonical_pack(b"test", blocksize=blocksize)
            async with Ignore(OverflowError, if_else=violation(problem)):
                await acanonical_pack(b"test", blocksize=blocksize)

    async def test_float_blocksize_is_invalid(self) -> None:
        problem = (  # fmt: skip
            "A float blocksize value didn't raise an `AttributeError`."
        )
        test_values = (-2.0, -1.0, 0.0, 1.0, 2.0)
        for blocksize in test_values:
            with Ignore(AttributeError, if_else=violation(problem)):
                canonical_pack(b"test", blocksize=blocksize)
            async with Ignore(AttributeError, if_else=violation(problem)):
                await acanonical_pack(b"test", blocksize=blocksize)

    async def test_minimum_size_given_by_item_count_declaration(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "An item count declaration which exceeds the count which is"
            "possible given the small size of the packing was allowed."
        )
        minimum_inputs = ((), (b"",), (b"", b""), (b"", b"", b""))
        tested_int_bytes = (1, 4, 8)
        for data in minimum_inputs:
            for int_bytes in tested_int_bytes:
                test = canonical_pack(*data, int_bytes=int_bytes)
                with Ignore(
                    CanonicalEncodingError, if_else=violation(problem)
                ):
                    canonical_unpack(
                        (test[0] + 1).to_bytes(1, BIG) + test[1:]
                    )
                test = canonical_pack(*data, int_bytes=int_bytes)
                async with Ignore(
                    CanonicalEncodingError, if_else=violation(problem)
                ):
                    await acanonical_unpack(
                        (test[0] + 1).to_bytes(1, BIG) + test[1:]
                    )


async def test_canonical_packs() -> None:
    PACK_PAD_INDEX = 33

    TEST_KEYS = [token_bytes(64) for _ in range(4)]
    HASHERS = (sha3_256, sha3_512, shake_128, shake_256)

    DEFAULT_PAD = b"\x00"
    DEFAULT_BLOCKSIZE = 1
    DEFAULT_INT_BYTES = 8

    DEFAULT_PACKING = canonical_pack()
    DEFAULT_ASYNC_PACKING = await acanonical_pack()

    # the default packing is the same for async & sync
    assert DEFAULT_PACKING == DEFAULT_ASYNC_PACKING
    assert DEFAULT_PACKING == await acanonical_pack(
        pad=DEFAULT_PAD,
        blocksize=DEFAULT_BLOCKSIZE,
        int_bytes=DEFAULT_INT_BYTES,
    )

    # the options can be used together
    assert canonical_unpack(
        canonical_pack(b"", pad=b"0", blocksize=127, int_bytes=16)
    ) == await acanonical_unpack(
        await acanonical_pack(b"", pad=b"0", blocksize=127, int_bytes=16)
    )

    test_inputs = [
        [token_bytes(token_bits(8)) for _ in range(2 + token_bits(2))]
        for _ in range(4)
    ]
    test_int_bytes = [1, 2, 3, 4]
    test_blocksizes = [
        SHA3_256_BLOCKSIZE,
        SHA3_512_BLOCKSIZE,
        SHAKE_128_BLOCKSIZE,
        SHAKE_256_BLOCKSIZE,
    ]
    test_pads = [b"\x01", b"\x80", b"\xff"]

    for inputs in test_inputs:
        # similar inputs do no produce the same output
        for int_bytes in test_int_bytes:
            result = canonical_pack(*inputs)
            aresult = await acanonical_pack(*inputs, int_bytes=int_bytes)
            assert result != aresult

            # but the items are still packed correctly & interoperably
            assert inputs == list(canonical_unpack(aresult))
            assert inputs == list(await acanonical_unpack(result))
        for blocksize in test_blocksizes:
            result = canonical_pack(*inputs)
            aresult = await acanonical_pack(*inputs, blocksize=blocksize)
            assert result != aresult
            assert 0 == len(aresult) % blocksize

            # but the items are still packed correctly & interoperably
            assert inputs == list(canonical_unpack(aresult))
            assert inputs == list(await acanonical_unpack(result))
        for pad, hasher, key in zip(test_pads, HASHERS, TEST_KEYS):
            result = canonical_pack(*inputs)
            aresult = await acanonical_pack(*inputs, pad=pad)
            assert result != aresult

            # but the items are still packed correctly & interoperably
            assert inputs == list(canonical_unpack(aresult))
            assert inputs == list(await acanonical_unpack(result))

            if not pad:
                continue

            # and the items & pads are used correctly in hashers
            obj = hasher()
            keyed_obj = obj.copy()
            digest_size = obj.digest_size
            encoded_key = encode_key(key, obj.block_size, pad=pad)
            packing = canonical_pack(
                (digest_size if digest_size else 64).to_bytes(
                    DEFAULT_INT_BYTES, BIG
                ),
                *inputs,
                blocksize=obj.block_size,
                pad=pad,
            )
            keyed_obj.update(encoded_key + packing)
            obj.update(packing)
            if digest_size:
                # keyed hashing of packed items works as expected
                assert keyed_obj.digest() == hash_bytes(
                    *inputs, pad=pad, key=key, hasher=hasher
                )
                assert keyed_obj.digest() == await ahash_bytes(
                    *inputs, pad=pad, key=key, hasher=hasher
                )

                # un-keyed hashing of packed items works as expected
                assert obj.digest() == hash_bytes(
                    *inputs, pad=pad, hasher=hasher
                )
                assert obj.digest() == await ahash_bytes(
                    *inputs, pad=pad, hasher=hasher
                )
            else:
                # keyed hashing of packed items works as expected
                assert keyed_obj.digest(64) == hash_bytes(
                    *inputs, pad=pad, key=key, hasher=hasher, size=64
                )
                assert keyed_obj.digest(64) == await ahash_bytes(
                    *inputs, pad=pad, key=key, hasher=hasher, size=64
                )

                # un-keyed hashing of packed items works as expected
                assert obj.digest(64) == hash_bytes(
                    *inputs, pad=pad, hasher=hasher, size=64
                )
                assert obj.digest(64) == await ahash_bytes(
                    *inputs, pad=pad, hasher=hasher, size=64
                )

        # the same inputs produce the same outputs
        for int_bytes in test_int_bytes:
            result = canonical_pack(*inputs, int_bytes=int_bytes)
            aresult = await acanonical_pack(*inputs, int_bytes=int_bytes)
            assert result == aresult

            # the relative location of the default pad declaration is
            # dependent on the size of integers used to represent item
            # lengths
            assert type(result) is bytes
            assert result[4 * int_bytes + 1] == DEFAULT_PAD[0]
        for blocksize in test_blocksizes:
            result = canonical_pack(*inputs, blocksize=blocksize)
            aresult = await acanonical_pack(*inputs, blocksize=blocksize)
            assert result == aresult
            assert 0 == len(result) % blocksize
        for pad in test_pads:
            result = canonical_pack(*inputs, pad=pad)
            aresult = await acanonical_pack(*inputs, pad=pad)
            assert result == aresult

            # the default integer size of 8 bytes puts the pad item at
            # the default location
            assert (result[PACK_PAD_INDEX] == pad[0]) if pad else 1

            # the relative location of the pad declaration is dependent
            # on the size of integers used to represent item lengths
            for int_bytes in test_int_bytes:
                result = canonical_pack(
                    *inputs, pad=pad, int_bytes=int_bytes
                )
                aresult = await acanonical_pack(
                    *inputs, pad=pad, int_bytes=int_bytes
                )
                assert result == aresult
                assert (result[4 * int_bytes + 1] == pad[0]) if pad else 1

    pad = b"Z"
    items = (
        b"testing",
        b"pad",
        b"character",
        b"location",
        b"in",
        b"result",
    )
    packing = bytearray(canonical_pack(*items, pad=pad))
    assert packing[PACK_PAD_INDEX] == pad[0]


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Initializer for the `Passcrypt` proof object.
"""

__all__ = ["PasscryptSession"]


import math
from hashlib import shake_128

from aiootp._typing import Typing as t
from aiootp._constants import B_TO_MB_RATIO, DEFAULT_AAD
from aiootp._constants import INT_BYTES, BIG, SHAKE_128_BLOCKSIZE
from aiootp.commons import FrozenInstance
from aiootp.generics import Domains, hash_bytes


class PasscryptSession(FrozenInstance):
    """
    Hanldes the initialization of running the `Passcrypt` hashing
    algorithm with sets of given user parameters.
    """

    __slots__ = (
        "passphrase",
        "salt",
        "aad",
        "mb",
        "cpu",
        "cores",
        "tag_size",
        "row_size",
        "rows",
        "total_size",
        "ram",
        "proof",
        "config",
    )

    _new_passcrypt_proof_kdf: t.Callable[[], t.XOFType] = shake_128(
        Domains.encode_constant(
            b"passcrypt_proof_kdf_salt",
            domain=Domains.PASSCRYPT,
            size=SHAKE_128_BLOCKSIZE,
        )
    ).copy

    def __init__(
        self,
        passphrase: bytes,
        salt: bytes,
        *,
        mb: int,
        cpu: int,
        cores: int,
        tag_size: int,
        config: t.ConfigType,
        aad: bytes = DEFAULT_AAD,
    ) -> None:
        """
        Efficiently stores user parameters.
        """
        self.config = config
        config.validate_inputs(passphrase=passphrase, salt=salt, aad=aad)
        config.validate_settings(
            mb=mb,
            cpu=cpu,
            cores=cores,
            tag_size=tag_size,
            salt_size=len(salt),
        )
        self.passphrase = passphrase
        self.salt = salt
        self.aad = aad
        self.mb = mb
        self.cpu = cpu
        self.cores = cores
        self.tag_size = tag_size

    def __iter__(
        self,
    ) -> t.Generator[t.Union[bytearray, t.Callable, int], None, None]:
        """
        Dumps the set of relevant parameters & function pointers for the
        `Passcrypt` worker. Gives the session a cleaner interface.
        """
        yield from (
            self.ram,
            self.proof.update,
            self.proof.digest,
            self.row_size,
            self.total_size,
        )

    def _hash_session_parameters(self) -> bytes:
        """
        Returns a 336-byte hash of the session's canonically encoded
        parameters.
        """
        return hash_bytes(
            Domains.PASSCRYPT,
            self.config.PACKED_METADATA,
            self.salt,
            self.aad,
            self.mb.to_bytes(INT_BYTES, BIG),
            self.cpu.to_bytes(INT_BYTES, BIG),
            self.cores.to_bytes(INT_BYTES, BIG),
            self.tag_size.to_bytes(INT_BYTES, BIG),
            key=self.passphrase + self.salt + self.aad,
            hasher=shake_128,
            size=336,
            pad=self.config.PASSCRYPT_PAD,
        )

    def prepare_session(self) -> t.Self:
        """
        Canonically hash the parameters to the function & calculate the
        dimensionality of the cache from the given settings.
        """
        rounds = max(
            (1, self.cpu // self.config.CPU_TO_DIGEST_PAIRS_PER_ROW_RATIO)
        )
        self.rows = math.ceil((B_TO_MB_RATIO * self.mb) / (336 * rounds))
        self.row_size = 336 * rounds
        self.total_size = self.row_size * self.rows
        parameters = self._hash_session_parameters()
        self.proof = self._new_passcrypt_proof_kdf()
        self.proof.update(parameters)
        return self

    def allocate_ram(self) -> t.Self:
        """
        Builds a virtual 2d memory cache out of a 1d bytearray to do
        efficient & in-place memory overwrites of segments of the cache
        with new proofs-of-work as the `Passcrypt` algorithm runs to
        completion.

        The bytearray is traversed to simulate the dimensionality of a
        columns=2*rounds, rows=ceil((1024*1024*mb) / (2*168*rounds)),
        2d array, where the unit measure for the width of one column is
        168-bytes (one digest from the `shake_128` `proof` object), &
        rounds=max([1, cpu // 2]).

        This procedure is designed to build the initial cache as fast as
        possible using the C implementation of `hashlib.shake_128` to
        better equalize the execution time between users & their
        adversaries. Quickly building the initial cache to the full size
        of the desired `mb` memory cost is also intended to reduce the
        inefficiencies of doing any resizing of the cache once the
        algorithm begins. This too is the main motivating factor for the
        size of each row being an equal multiple of 336, as it allows
        cache traversal & insertions without needing to plan separately
        for how to treat insertions once the end of a row is reached.
        """
        self.ram = bytearray()
        size = self.total_size
        # fmt: off
        max_size = (B_TO_MB_RATIO * 512) - 1  # 512MiB, max digest size
        while size > max_size:                # of shake_128 in python
            self.ram.extend(self.proof.digest(max_size))
            self.proof.update(self.ram[-168:])
            size -= max_size
        # fmt: on
        if size:
            self.ram.extend(self.proof.digest(size))
            self.proof.update(self.ram[-168:])
        return self


module_api = dict(
    PasscryptSession=t.add_type(PasscryptSession),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)
